from typing import Any, Generic, Literal, TypeAlias, TypeVar

from xdsl.dialects.builtin import (
    ArrayAttr,
    DictionaryAttr,
    FloatData,
    IndexType,
    IntAttr,
    IntegerAttr,
    IntegerType,
)
from xdsl.ir import Attribute, ParametrizedAttribute
from xdsl.irdl import ParameterDef, irdl_attr_definition
from xdsl.utils.hints import isa


class Class1:
    pass


class SubClass1(Class1):
    pass


class Class2:
    pass


################################################################################
# Any
################################################################################


def test_any_hint():
    """Test that the we can check if a value is satisfying `Any`."""
    assert isa(3, Any)
    assert isa([], Any)
    assert isa([2], Any)
    assert isa(int, Any)


################################################################################
# Class
################################################################################


def test_class_hint_correct():
    """Test that a class hint is being satisfied by giving a class instance."""
    assert isa(Class1(), Class1)
    assert isa(SubClass1(), SubClass1)
    assert isa(Class2(), Class2)
    assert isa(True, bool)
    assert isa(10, int)
    assert isa("", str)


def test_class_hint_class():
    """Test that a class hint is not being satisfied by the class object."""
    assert not isa(Class1, Class1)


def test_class_hint_wrong_instance():
    """
    Test that a class hint is not being satisfied by another class instance.
    """
    assert not isa(Class2(), Class1)
    assert not isa(0, Class1)
    assert not isa("", Class1)
    assert not isa("", int)


def test_class_hint_subclass():
    """
    Test that a class hint is satisfied by a subclass.
    """
    assert isa(SubClass1(), Class1)
    assert isa(True, int)  # bool is a subclass of int


################################################################################
# List
################################################################################


def test_list_hint_empty():
    """Test that empty lists satisfy all list hints."""
    assert isa([], list[int])
    assert isa([], list[bool])
    assert isa([], list[Class1])


def test_list_hint_correct():
    """
    Test that list hints work correcly on non-empty lists of the right type.
    """
    assert isa([42], list[int])
    assert isa([0, 3, 5], list[int])
    assert isa([False], list[bool])
    assert isa([True, False], list[bool])
    assert isa([Class1(), SubClass1()], list[Class1])


def test_list_hint_not_list_failure():
    """Test that list hints work correcly on non lists."""
    assert not isa(0, list[int])
    assert not isa(0, list[Any])
    assert not isa(True, list[bool])
    assert not isa(True, list[Any])
    assert not isa("", list[Any])
    assert not isa("", list[str])
    assert not isa(Class1(), list[Class1])
    assert not isa(Class1(), list[Any])
    assert not isa({}, list[dict[Any, Any]])
    assert not isa({}, list[Any])


def test_list_hint_failure():
    """
    Test that list hints work correcly on non-empty lists of the wrong type.
    """
    assert not isa([0], list[bool])
    assert not isa([0, True], list[bool])
    assert not isa([True, 0], list[bool])
    assert not isa([True, False, True, 0], list[bool])
    assert not isa([Class2()], list[Class1])


def test_list_hint_nested():
    """
    Test that we can check nested list hints.
    """
    assert isa([[]], list[list[int]])
    assert isa([[0]], list[list[int]])
    assert isa([[0], [2, 1]], list[list[int]])
    assert not isa([[0], [2, 1], [""]], list[list[int]])
    assert not isa([[0], [2, 1], [32, ""]], list[list[int]])
    assert isa([], list[list[list[int]]])
    assert isa([[]], list[list[list[int]]])
    assert isa([[[]]], list[list[list[int]]])
    assert isa([[], [[]]], list[list[list[int]]])
    assert isa([[], [[0, 32]]], list[list[list[int]]])


################################################################################
# Tuple
################################################################################


def test_tuple_hint_empty():
    """Test that empty tuple satisfy all tuple hints."""
    assert isa(tuple(), tuple[int, ...])
    assert isa(tuple(), tuple[bool, ...])
    assert isa(tuple(), tuple[Class1, ...])


def test_tuple_hint_correct():
    """
    Test that tuple hints work correcly on non-empty tuples of the right type.
    """
    assert isa((42,), tuple[int])
    assert isa((0, 3, 5), tuple[int, int, int])
    assert isa((False,), tuple[bool])
    assert isa((True, False), tuple[bool, ...])
    assert isa((True, 1, "test"), tuple[bool, int, str])
    assert isa((Class1(), SubClass1()), tuple[Class1, ...])


def test_tuple_hint_not_list_failure():
    """Test that tuple hints work correcly on non tuple."""
    assert not isa(0, tuple[int])
    assert not isa(0, tuple[Any])
    assert not isa(True, tuple[bool])
    assert not isa(True, tuple[Any])
    assert not isa("", tuple[Any])
    assert not isa("", tuple[str])
    assert not isa(Class1(), tuple[Class1])
    assert not isa(Class1(), tuple[Any])
    assert not isa({}, tuple[dict[Any, Any]])
    assert not isa({}, tuple[Any])


def test_tuple_hint_failure():
    """
    Test that tuple hints work correcly on non-empty tuples of the wrong type.
    """
    assert not isa((0,), tuple[bool])
    assert not isa((0, True), tuple[bool, bool])
    assert not isa((0, True), tuple[int])
    assert not isa((True, 0), tuple[bool, bool])
    assert not isa((True, False, True, 0), tuple[bool, ...])
    assert not isa((Class2(),), tuple[Class1])


def test_tuple_hint_nested():
    """
    Test that we can check nested tuple hints.
    """
    assert isa(((),), tuple[tuple[int, ...]])
    assert isa(((0,),), tuple[tuple[int]])
    assert isa((0, (1, 2)), tuple[int, tuple[int, int]])
    assert isa(((0, 1), (2, 3), (4, 5)), tuple[tuple[int, int], ...])
    assert not isa(((0, 1), (2, 3), (4, "5")), tuple[tuple[int, int], ...])
    assert not isa(((0, 1), (2, 3), (4, "5")), tuple[tuple[int, ...], ...])


################################################################################
# Set
################################################################################


def test_set_hint_empty():
    """Test that empty set satisfy all set hints."""
    assert isa(set(), set[int])
    assert isa(set(), set[bool])
    assert isa(set(), set[Class1])


def test_set_hint_correct():
    """
    Test that set hints work correcly on non-empty sets of the right type.
    """
    assert isa({42}, set[int])
    assert isa({0, 3, 5}, set[int])
    assert isa({False}, set[bool])
    assert isa({True, False}, set[bool])
    assert isa({True, 1, "test"}, set[bool | int | str])
    assert isa({Class1(), SubClass1()}, set[Class1])


def test_set_hint_not_list_failure():
    """Test that set hints work correcly on non set."""
    assert not isa(0, set[int])
    assert not isa(0, set[Any])
    assert not isa(True, set[bool])
    assert not isa(True, set[Any])
    assert not isa("", set[Any])
    assert not isa("", set[str])
    assert not isa(Class1(), set[Class1])
    assert not isa(Class1(), set[Any])
    assert not isa([], set[dict[Any, Any]])
    assert not isa([], set[Any])


def test_set_hint_failure():
    """
    Test that set hints work correcly on non-empty sets of the wrong type.
    """
    assert not isa({0}, set[bool])
    assert not isa({0, "hello"}, set[int])


################################################################################
# Dictionary
################################################################################


def test_dict_hint_empty():
    """Test that empty dicts satisfy all dict hints."""
    assert isa({}, dict[int, bool])
    assert isa({}, dict[Any, Any])
    assert isa({}, dict[Class1, Any])
    assert isa({}, dict[str, int])


def test_dict_hint_correct():
    """
    Test that dict hints work correcly on non-empty dicts of the right type.
    """
    assert isa({"": 0}, dict[str, int])
    assert isa({"": 0, "a": 32}, dict[str, int])
    assert isa({0: "", 32: "a"}, dict[int, str])
    assert isa({True: Class1(), False: SubClass1()}, dict[bool, Class1])


def test_dict_hint_not_dict_failure():
    """Test that dict hints work correcly on non dicts."""
    assert not isa(0, dict[int, int])
    assert not isa(0, dict[Any, Any])
    assert not isa(True, dict[bool, bool])
    assert not isa(True, dict[Any, Any])
    assert not isa("", dict[Any, Any])
    assert not isa(Class1(), dict[Any, Any])
    assert not isa([], dict[Any, Any])


def test_dict_hint_failure():
    """
    Test that dict hints work correcly on non-empty dicts of the wrong type.
    """
    assert not isa({"": ""}, dict[int, str])
    assert not isa({0: 0}, dict[int, str])
    assert not isa({0: "0", 2: 2}, dict[int, str])
    assert not isa({0: "0", 2: "1", "2": "2"}, dict[int, str])


threeDict: TypeAlias = dict[int, dict[int, dict[int, str]]]


def test_dict_hint_nested():
    """
    Test that we can check nested dict hints.
    """
    assert isa({}, dict[int, dict[int, str]])
    assert isa({0: {}}, dict[int, dict[int, str]])
    assert isa({0: {1: ""}}, dict[int, dict[int, str]])

    assert not isa({"": {}}, dict[int, dict[int, str]])
    assert not isa({0: ""}, dict[int, dict[int, str]])
    assert not isa({0: {"": ""}}, dict[int, dict[int, str]])
    assert not isa({0: {0: 0}}, dict[int, dict[int, str]])

    assert isa({}, threeDict)
    assert isa({0: {}}, threeDict)
    assert isa({0: {0: {}}}, threeDict)
    assert isa({0: {}, 1: {0: {}}}, threeDict)
    assert isa({0: {}, 1: {0: {0: "0", 1: "32"}}}, threeDict)


################################################################################
# GenericData
################################################################################


def test_generic_data():
    attr = ArrayAttr([IntAttr(0)])

    assert isa(attr, ArrayAttr[IntAttr])
    assert isa(attr, ArrayAttr[IntAttr | FloatData])
    assert not isa(attr, ArrayAttr[FloatData])

    attr2 = ArrayAttr([IntAttr(0), FloatData(0.0)])

    assert not isa(attr2, ArrayAttr[IntAttr])
    assert isa(attr2, ArrayAttr[IntAttr | FloatData])
    assert not isa(attr2, ArrayAttr[FloatData])

    intattr = IntAttr(42)

    assert not isa(intattr, ArrayAttr[Attribute])
    assert not isa(intattr, DictionaryAttr)

    integerattr = IntegerAttr.from_index_int_value(42)

    assert not isa(integerattr, ArrayAttr[Attribute])
    assert not isa(integerattr, DictionaryAttr)


def test_nested_generic_data():
    attr = ArrayAttr([IntegerAttr.from_index_int_value(0)])

    assert isa(attr, ArrayAttr[Attribute])
    assert isa(attr, ArrayAttr[IntegerAttr[IndexType]])
    assert isa(attr, ArrayAttr[IntegerAttr[IndexType | IntegerType]])
    assert not isa(attr, ArrayAttr[IntegerAttr[IntegerType]])


################################################################################
# ParametrizedAttribute
################################################################################

_T = TypeVar("_T", bound=Attribute)


@irdl_attr_definition
class MyParamAttr(Generic[_T], ParametrizedAttribute):
    name = "test.param"

    v: ParameterDef[_T]


def test_parametrized_attribute():
    attr = MyParamAttr[IntAttr]([IntAttr(0)])

    assert isa(attr, MyParamAttr)
    assert isa(attr, MyParamAttr[IntAttr])
    assert isa(attr, MyParamAttr[IntAttr | FloatData])
    assert not isa(attr, MyParamAttr[FloatData])


################################################################################
# Literal
################################################################################


def test_literal():
    assert isa("string", Literal["string"])
    assert isa("string", Literal["string", "another string"])
    assert isa("another string", Literal["string", "another string"])
    assert not isa("not this string", Literal["string", "another string"])

    assert isa(1, Literal[1])
    assert isa(1, Literal[1, 2])
    assert isa(2, Literal[1, 2])
    assert not isa(3, Literal[1, 2])

    assert not isa(1, Literal["1"])
    assert not isa("1", Literal[1])



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator, Mapping, Sequence
from dataclasses import dataclass
from enum import Enum
from math import prod
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Generic,
    TypeAlias,
    TypeVar,
    cast,
    overload,
)

from typing_extensions import Self

from xdsl.ir import (
    Attribute,
    AttributeCovT,
    AttributeInvT,
    Block,
    BlockOps,
    Data,
    Dialect,
    Operation,
    ParametrizedAttribute,
    Region,
    SSAValue,
    TypeAttribute,
    TypedAttribute,
)
from xdsl.ir.affine import AffineMap, AffineSet
from xdsl.irdl import (
    AllOf,
    AnyAttr,
    AnyOf,
    AttrConstraint,
    GenericData,
    IRDLOperation,
    MessageConstraint,
    ParamAttrConstraint,
    ParameterDef,
    VarOperand,
    VarOpResult,
    attr_constr_coercion,
    irdl_attr_definition,
    irdl_op_definition,
    irdl_to_attr_constraint,
    opt_attr_def,
    region_def,
    var_operand_def,
    var_result_def,
)
from xdsl.traits import (
    IsolatedFromAbove,
    NoTerminator,
    OptionalSymbolOpInterface,
    SymbolTable,
)
from xdsl.utils.deprecation import deprecated_constructor
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.hints import isa

if TYPE_CHECKING:
    from xdsl.parser import AttrParser, Parser
    from xdsl.printer import Printer


class ShapedType(ABC):
    @abstractmethod
    def get_num_dims(self) -> int: ...

    @abstractmethod
    def get_shape(self) -> tuple[int, ...]: ...

    def element_count(self) -> int:
        return prod(self.get_shape())


class AnyShapedType(AttrConstraint):
    def verify(self, attr: Attribute, constraint_vars: dict[str, Attribute]) -> None:
        if not isinstance(attr, ShapedType):
            raise Exception(f"expected type ShapedType but got {attr}")


_ContainerElementTypeT = TypeVar(
    "_ContainerElementTypeT", bound=Attribute | None, covariant=True
)


class ContainerType(Generic[_ContainerElementTypeT], ABC):
    @abstractmethod
    def get_element_type(self) -> _ContainerElementTypeT:
        pass


@irdl_attr_definition
class NoneAttr(ParametrizedAttribute):
    """An attribute representing the absence of an attribute."""

    name = "none"


@dataclass(frozen=True)
class ArrayOfConstraint(AttrConstraint):
    """
    A constraint that enforces an ArrayData whose elements all satisfy
    the elem_constr.
    """

    elem_constr: AttrConstraint

    def __init__(self, constr: Attribute | type[Attribute] | AttrConstraint):
        object.__setattr__(self, "elem_constr", attr_constr_coercion(constr))

    def verify(self, attr: Attribute, constraint_vars: dict[str, Attribute]) -> None:
        if not isinstance(attr, ArrayAttr):
            raise VerifyException(f"expected ArrayData attribute, but got {attr}")
        for e in cast(ArrayAttr[Attribute], attr).data:
            self.elem_constr.verify(e, constraint_vars)


@irdl_attr_definition
class ArrayAttr(GenericData[tuple[AttributeCovT, ...]], Iterable[AttributeCovT]):
    name = "array"

    def __init__(self, param: Iterable[AttributeCovT]) -> None:
        super().__init__(tuple(param))

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> tuple[AttributeCovT, ...]:
        with parser.in_angle_brackets():
            data = parser.parse_comma_separated_list(
                parser.Delimiter.SQUARE, parser.parse_attribute
            )
            # the type system can't ensure that the elements are of type _ArrayAttrT
            result = cast(tuple[AttributeCovT, ...], tuple(data))
            return result

    def print_parameter(self, printer: Printer) -> None:
        printer.print_string("[")
        printer.print_list(self.data, printer.print_attribute)
        printer.print_string("]")

    @staticmethod
    def generic_constraint_coercion(args: tuple[Any]) -> AttrConstraint:
        if len(args) == 1:
            return ArrayOfConstraint(irdl_to_attr_constraint(args[0]))
        if len(args) == 0:
            return ArrayOfConstraint(AnyAttr())
        raise TypeError(
            f"Attribute ArrayAttr expects at most 1 type"
            f" parameter, but {len(args)} were given"
        )

    def verify(self) -> None:
        for idx, val in enumerate(self.data):
            if not isinstance(val, Attribute):
                raise VerifyException(
                    f"{self.name} data expects attribute list, but {idx} "
                    f"element is of type {type(val)}"
                )

    def __len__(self):
        return len(self.data)

    def __iter__(self) -> Iterator[AttributeCovT]:
        return iter(self.data)


AnyArrayAttr: TypeAlias = ArrayAttr[Attribute]


@irdl_attr_definition
class StringAttr(Data[str]):
    name = "string"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> str:
        with parser.in_angle_brackets():
            return parser.parse_str_literal()

    def print_parameter(self, printer: Printer) -> None:
        printer.print_string(f'"{self.data}"')


@irdl_attr_definition
class BytesAttr(Data[bytes]):
    name = "bytes"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> bytes:
        with parser.in_angle_brackets():
            return parser.parse_bytes_literal()

    def print_parameter(self, printer: Printer) -> None:
        printer.print_string(f'"{self.data}"')


@irdl_attr_definition
class SymbolNameAttr(ParametrizedAttribute):
    name = "symbol_name"
    data: ParameterDef[StringAttr]

    def __init__(self, data: str | StringAttr) -> None:
        if isinstance(data, str):
            data = StringAttr(data)
        super().__init__([data])


@irdl_attr_definition
class SymbolRefAttr(ParametrizedAttribute):
    name = "symbol_ref"
    root_reference: ParameterDef[StringAttr]
    nested_references: ParameterDef[ArrayAttr[StringAttr]]

    def __init__(
        self,
        root: str | StringAttr,
        nested: Sequence[str] | Sequence[StringAttr] | ArrayAttr[StringAttr] = [],
    ) -> None:
        if isinstance(root, str):
            root = StringAttr(root)
        if not isinstance(nested, ArrayAttr):
            nested = ArrayAttr(
                [StringAttr(x) if isinstance(x, str) else x for x in nested]
            )
        super().__init__([root, nested])

    def string_value(self):
        root = self.root_reference.data
        for ref in self.nested_references.data:
            root += "." + ref.data
        return root


class EmptyArrayAttrConstraint(AttrConstraint):
    """
    Constrain attribute to be empty ArrayData
    """

    def verify(self, attr: Attribute, constraint_vars: dict[str, Attribute]) -> None:
        if not isinstance(attr, ArrayAttr):
            raise VerifyException(f"expected ArrayData attribute, but got {attr}")
        attr = cast(ArrayAttr[Attribute], attr)
        if attr.data:
            raise VerifyException(f"expected empty array, but got {attr}")


FlatSymbolRefAttrConstraint = MessageConstraint(
    ParamAttrConstraint(SymbolRefAttr, [AnyAttr(), EmptyArrayAttrConstraint()]),
    "Unexpected nested symbols in FlatSymbolRefAttr.",
)
"""Constrain SymbolRef to be FlatSymbolRef"""

FlatSymbolRefAttr = Annotated[SymbolRefAttr, FlatSymbolRefAttrConstraint]
"""SymbolRef constrained to have an empty `nested_references` property."""


@irdl_attr_definition
class IntAttr(Data[int]):
    name = "builtin.int"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> int:
        with parser.in_angle_brackets():
            data = parser.parse_integer()
            return data

    def print_parameter(self, printer: Printer) -> None:
        with printer.in_angle_brackets():
            printer.print_string(f"{self.data}")


class Signedness(Enum):
    "Signedness semantics for integer"

    SIGNLESS = 0
    "No signedness semantics"

    SIGNED = 1
    UNSIGNED = 2

    def value_range(self, bitwidth: int) -> tuple[int, int]:
        """
        For a given bitwidth, returns (min, max+1), where min and max are the smallest and
        largest representable values.

        Signless integers are bit patterns, so the representable range is the union of the
        signed and unsigned representable ranges.
        """
        match self:
            case Signedness.SIGNLESS:
                min_value = -((1 << bitwidth) >> 1)
                max_value = 1 << bitwidth
            case Signedness.SIGNED:
                min_value = -((1 << bitwidth) >> 1)
                max_value = 1 << (bitwidth - 1)
            case Signedness.UNSIGNED:
                min_value = 0
                max_value = 1 << bitwidth

        return min_value, max_value


@irdl_attr_definition
class SignednessAttr(Data[Signedness]):
    name = "builtin.signedness"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> Signedness:
        with parser.in_angle_brackets():
            if parser.parse_optional_keyword("signless") is not None:
                return Signedness.SIGNLESS
            if parser.parse_optional_keyword("signed") is not None:
                return Signedness.SIGNED
            if parser.parse_optional_keyword("unsigned") is not None:
                return Signedness.UNSIGNED
            parser.raise_error("`signless`, `signed`, or `unsigned` expected")

    def print_parameter(self, printer: Printer) -> None:
        with printer.in_angle_brackets():
            data = self.data
            if data == Signedness.SIGNLESS:
                printer.print_string("signless")
            elif data == Signedness.SIGNED:
                printer.print_string("signed")
            elif data == Signedness.UNSIGNED:
                printer.print_string("unsigned")
            else:
                raise ValueError(f"Invalid signedness {data}")


@irdl_attr_definition
class IntegerType(ParametrizedAttribute, TypeAttribute):
    name = "integer_type"
    width: ParameterDef[IntAttr]
    signedness: ParameterDef[SignednessAttr]

    def __init__(
        self,
        data: int | IntAttr,
        signedness: Signedness | SignednessAttr = Signedness.SIGNLESS,
    ) -> None:
        if isinstance(data, int):
            data = IntAttr(data)
        if isinstance(signedness, Signedness):
            signedness = SignednessAttr(signedness)
        super().__init__([data, signedness])

    def value_range(self) -> tuple[int, int]:
        return self.signedness.data.value_range(self.width.data)


i64 = IntegerType(64)
i32 = IntegerType(32)
i1 = IntegerType(1)


SignlessIntegerConstraint = ParamAttrConstraint(
    IntegerType, [IntAttr, SignednessAttr(Signedness.SIGNLESS)]
)
"""Type constraint for signless IntegerType."""

AnySignlessIntegerType: TypeAlias = Annotated[IntegerType, SignlessIntegerConstraint]
"""Type alias constrained to signless IntegerType."""


@irdl_attr_definition
class UnitAttr(ParametrizedAttribute):
    name = "unit"


@irdl_attr_definition
class LocationAttr(ParametrizedAttribute):
    """
    An attribute representing source code location.
    Only supports unknown locations for now.
    """

    name = "loc"


@irdl_attr_definition
class IndexType(ParametrizedAttribute):
    name = "index"


_IntegerAttrType = TypeVar(
    "_IntegerAttrType", bound=IntegerType | IndexType, covariant=True
)
AnySignlessIntegerOrIndexType: TypeAlias = Annotated[
    Attribute, AnyOf([IndexType, SignlessIntegerConstraint])
]
"""Type alias constrained to IndexType or signless IntegerType."""


@irdl_attr_definition
class IntegerAttr(
    Generic[_IntegerAttrType],
    TypedAttribute[_IntegerAttrType],
):
    name = "integer"
    value: ParameterDef[IntAttr]
    type: ParameterDef[_IntegerAttrType]

    @overload
    def __init__(
        self: IntegerAttr[_IntegerAttrType],
        value: int | IntAttr,
        value_type: _IntegerAttrType,
    ) -> None: ...

    @overload
    def __init__(
        self: IntegerAttr[IntegerType], value: int | IntAttr, value_type: int
    ) -> None: ...

    def __init__(
        self, value: int | IntAttr, value_type: int | IntegerType | IndexType
    ) -> None:
        if isinstance(value, int):
            value = IntAttr(value)
        if isinstance(value_type, int):
            value_type = IntegerType(value_type)
        super().__init__([value, value_type])

    @staticmethod
    def from_int_and_width(value: int, width: int) -> IntegerAttr[IntegerType]:
        return IntegerAttr(value, width)

    @staticmethod
    def from_index_int_value(value: int) -> IntegerAttr[IndexType]:
        return IntegerAttr(value, IndexType())

    def verify(self) -> None:
        if isinstance(int_type := self.type, IndexType):
            return

        min_value, max_value = int_type.value_range()

        if not (min_value <= self.value.data < max_value):
            raise VerifyException(
                f"Integer value {self.value.data} is out of range for "
                f"type {self.type} which supports values in the "
                f"range [{min_value}, {max_value})"
            )

    @classmethod
    def parse_with_type(
        cls: type[IntegerAttr[_IntegerAttrType]],
        parser: AttrParser,
        type: Attribute,
    ) -> IntegerAttr[_IntegerAttrType]:
        assert isinstance(type, IntegerType) or isinstance(type, IndexType)
        return cast(
            IntegerAttr[_IntegerAttrType],
            IntegerAttr(parser.parse_integer(allow_boolean=(type == i1)), type),
        )

    def print_without_type(self, printer: Printer):
        return printer.print(self.value.data)


AnyIntegerAttr: TypeAlias = IntegerAttr[IntegerType | IndexType]
BoolAttr: TypeAlias = IntegerAttr[Annotated[IntegerType, IntegerType(1)]]


class _FloatType(ABC):
    @property
    @abstractmethod
    def get_bitwidth(self) -> int:
        raise NotImplementedError()


@irdl_attr_definition
class BFloat16Type(ParametrizedAttribute, TypeAttribute, _FloatType):
    name = "bf16"

    @property
    def get_bitwidth(self) -> int:
        return 16


@irdl_attr_definition
class Float16Type(ParametrizedAttribute, TypeAttribute, _FloatType):
    name = "f16"

    @property
    def get_bitwidth(self) -> int:
        return 16


@irdl_attr_definition
class Float32Type(ParametrizedAttribute, TypeAttribute, _FloatType):
    name = "f32"

    @property
    def get_bitwidth(self) -> int:
        return 32


@irdl_attr_definition
class Float64Type(ParametrizedAttribute, TypeAttribute, _FloatType):
    name = "f64"

    @property
    def get_bitwidth(self) -> int:
        return 64


@irdl_attr_definition
class Float80Type(ParametrizedAttribute, TypeAttribute, _FloatType):
    name = "f80"

    @property
    def get_bitwidth(self) -> int:
        return 80


@irdl_attr_definition
class Float128Type(ParametrizedAttribute, TypeAttribute, _FloatType):
    name = "f128"

    @property
    def get_bitwidth(self) -> int:
        return 128


AnyFloat: TypeAlias = (
    BFloat16Type | Float16Type | Float32Type | Float64Type | Float80Type | Float128Type
)


@irdl_attr_definition
class FloatData(Data[float]):
    name = "float_data"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> float:
        with parser.in_angle_brackets():
            return float(parser.parse_number())

    def print_parameter(self, printer: Printer) -> None:
        printer.print_string(f"{self.data}")


_FloatAttrType = TypeVar("_FloatAttrType", bound=AnyFloat, covariant=True)


@irdl_attr_definition
class FloatAttr(Generic[_FloatAttrType], ParametrizedAttribute):
    name = "float"

    value: ParameterDef[FloatData]
    type: ParameterDef[_FloatAttrType]

    @overload
    def __init__(self, data: float | FloatData, type: _FloatAttrType) -> None: ...

    @overload
    def __init__(self, data: float | FloatData, type: int) -> None: ...

    def __init__(
        self, data: float | FloatData, type: int | _FloatAttrType | AnyFloat
    ) -> None:
        if isinstance(data, FloatData):
            data_attr = data
        else:
            data_attr = FloatData(data)
        if isinstance(type, int):
            if type == 16:
                type = Float16Type()
            elif type == 32:
                type = Float32Type()
            elif type == 64:
                type = Float64Type()
            elif type == 80:
                type = Float80Type()
            elif type == 128:
                type = Float128Type()
            else:
                raise ValueError(f"Invalid bitwidth: {type}")
        super().__init__([data_attr, type])


AnyFloatAttr: TypeAlias = FloatAttr[AnyFloat]


@irdl_attr_definition
class ComplexType(ParametrizedAttribute, TypeAttribute):
    name = "complex"
    element_type: ParameterDef[IntegerType | AnyFloat]

    def __init__(self, element_type: IntegerType | AnyFloat) -> None:
        ParametrizedAttribute.__init__(self, [element_type])


@irdl_attr_definition
class DictionaryAttr(GenericData[dict[str, Attribute]]):
    name = "dictionary"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> dict[str, Attribute]:
        return parser.parse_optional_dictionary_attr_dict()

    def print_parameter(self, printer: Printer) -> None:
        printer.print_attr_dict(self.data)

    @staticmethod
    def generic_constraint_coercion(args: tuple[Any]) -> AttrConstraint:
        raise Exception(f"Unsupported operation on {DictionaryAttr.name}")

    def verify(self) -> None:
        return super().verify()


@irdl_attr_definition
class TupleType(ParametrizedAttribute):
    name = "tuple"

    types: ParameterDef[ArrayAttr[Attribute]]

    def __init__(self, types: list[Attribute] | ArrayAttr[Attribute]) -> None:
        if isinstance(types, list):
            types = ArrayAttr(types)
        super().__init__([types])


@irdl_attr_definition
class VectorType(
    Generic[AttributeCovT],
    ParametrizedAttribute,
    TypeAttribute,
    ShapedType,
    ContainerType[AttributeCovT],
):
    name = "vector"

    shape: ParameterDef[ArrayAttr[IntAttr]]
    element_type: ParameterDef[AttributeCovT]
    num_scalable_dims: ParameterDef[IntAttr]

    def __init__(
        self: VectorType[AttributeCovT],
        element_type: AttributeCovT,
        shape: Iterable[int | IntAttr],
        num_scalable_dims: int | IntAttr = 0,
    ) -> None:
        shape = ArrayAttr(
            [IntAttr(dim) if isinstance(dim, int) else dim for dim in shape]
        )
        if isinstance(num_scalable_dims, int):
            num_scalable_dims = IntAttr(num_scalable_dims)
        super().__init__([shape, element_type, num_scalable_dims])

    def get_num_dims(self) -> int:
        return len(self.shape.data)

    def get_num_scalable_dims(self) -> int:
        return self.num_scalable_dims.data

    def get_shape(self) -> tuple[int, ...]:
        return tuple(i.data for i in self.shape)

    def get_element_type(self) -> AttributeCovT:
        return self.element_type

    def verify(self):
        if self.get_num_scalable_dims() < 0:
            raise VerifyException(
                f"Number of scalable dimensions {self.get_num_dims()} cannot"
                " be negative"
            )
        if self.get_num_scalable_dims() > self.get_num_dims():
            raise VerifyException(
                f"Number of scalable dimensions {self.get_num_scalable_dims()}"
                " cannot be larger than number of dimensions"
                f" {self.get_num_dims()}"
            )

    @deprecated_constructor
    @staticmethod
    def from_element_type_and_shape(
        referenced_type: AttributeInvT,
        shape: Iterable[int | IntegerAttr[IndexType]],
        num_scalable_dims: int | IntAttr = 0,
    ) -> VectorType[AttributeInvT]:
        if isinstance(num_scalable_dims, int):
            num_scalable_dims = IntAttr(num_scalable_dims)
        shape_int = [
            IntAttr(dim) if isinstance(dim, int) else dim.value.data for dim in shape
        ]
        return VectorType(referenced_type, shape_int, num_scalable_dims)

    @deprecated_constructor
    @staticmethod
    def from_params(
        referenced_type: AttributeInvT,
        shape: ArrayAttr[IntegerAttr[IntegerType]] = ArrayAttr(
            [IntegerAttr.from_int_and_width(1, 64)]
        ),
        num_scalable_dims: IntAttr = IntAttr(0),
    ) -> VectorType[AttributeInvT]:
        shape_int = [dim.value.data for dim in shape.data]
        return VectorType(referenced_type, shape_int, num_scalable_dims)


AnyVectorType: TypeAlias = VectorType[Attribute]


@irdl_attr_definition
class TensorType(
    Generic[AttributeCovT],
    ParametrizedAttribute,
    TypeAttribute,
    ShapedType,
    ContainerType[AttributeCovT],
):
    name = "tensor"

    shape: ParameterDef[ArrayAttr[IntAttr]]
    element_type: ParameterDef[AttributeCovT]
    encoding: ParameterDef[Attribute]

    def __init__(
        self: TensorType[AttributeCovT],
        element_type: AttributeCovT,
        shape: Iterable[int | IntAttr],
        encoding: Attribute = NoneAttr(),
    ):
        shape = ArrayAttr(
            [IntAttr(dim) if isinstance(dim, int) else dim for dim in shape]
        )
        super().__init__([shape, element_type, encoding])

    def get_num_dims(self) -> int:
        return len(self.shape.data)

    def get_shape(self) -> tuple[int, ...]:
        return tuple(i.data for i in self.shape.data)

    def get_element_type(self) -> AttributeCovT:
        return self.element_type

    @deprecated_constructor
    @staticmethod
    def from_type_and_list(
        referenced_type: AttributeInvT,
        shape: Iterable[int | IntegerAttr[IndexType]] | None = None,
        encoding: Attribute = NoneAttr(),
    ) -> TensorType[AttributeInvT]:
        if shape is None:
            shape = [1]
        shape_int = [
            IntAttr(dim) if isinstance(dim, int) else dim.value.data for dim in shape
        ]
        return TensorType(referenced_type, shape_int, encoding)

    @deprecated_constructor
    @staticmethod
    def from_params(
        referenced_type: AttributeInvT,
        shape: AnyArrayAttr = AnyArrayAttr([IntegerAttr.from_int_and_width(1, 64)]),
        encoding: Attribute = NoneAttr(),
    ) -> TensorType[AttributeInvT]:
        if not isa(shape, ArrayAttr[AnyIntegerAttr]):
            raise TypeError(f"Unsupported shape type {type(shape)}")
        shape_int = [dim.value.data for dim in shape.data]
        return TensorType(referenced_type, shape_int, encoding)


AnyTensorType: TypeAlias = TensorType[Attribute]


@irdl_attr_definition
class UnrankedTensorType(Generic[AttributeCovT], ParametrizedAttribute, TypeAttribute):
    name = "unranked_tensor"

    element_type: ParameterDef[AttributeCovT]

    def __init__(
        self: UnrankedTensorType[AttributeCovT], element_type: AttributeCovT
    ) -> None:
        super().__init__([element_type])


AnyUnrankedTensorType: TypeAlias = UnrankedTensorType[Attribute]


@dataclass(frozen=True, init=False)
class ContainerOf(AttrConstraint):
    """A type constraint that can be nested once in a vector or a tensor."""

    elem_constr: AttrConstraint

    def __init__(
        self, elem_constr: Attribute | type[Attribute] | AttrConstraint
    ) -> None:
        object.__setattr__(self, "elem_constr", attr_constr_coercion(elem_constr))

    def verify(self, attr: Attribute, constraint_vars: dict[str, Attribute]) -> None:
        if isinstance(attr, VectorType) or isinstance(attr, TensorType):
            attr = cast(VectorType[Attribute] | TensorType[Attribute], attr)
            self.elem_constr.verify(attr.element_type, constraint_vars)
        else:
            self.elem_constr.verify(attr, constraint_vars)


VectorOrTensorOf: TypeAlias = (
    VectorType[AttributeCovT]
    | TensorType[AttributeCovT]
    | UnrankedTensorType[AttributeCovT]
)

RankedVectorOrTensorOf: TypeAlias = (
    VectorType[AttributeCovT] | TensorType[AttributeCovT]
)


@dataclass(frozen=True)
class VectorRankConstraint(AttrConstraint):
    """
    Constrain a vector to be of a given rank.
    """

    expected_rank: int
    """The expected vector rank."""

    def verify(self, attr: Attribute, constraint_vars: dict[str, Attribute]) -> None:
        if not isinstance(attr, VectorType):
            raise VerifyException(f"{attr} should be of type VectorType.")
        if attr.get_num_dims() != self.expected_rank:
            raise VerifyException(
                f"Expected vector rank to be {self.expected_rank}, got {attr.get_num_dims()}."
            )


@dataclass(frozen=True)
class VectorBaseTypeConstraint(AttrConstraint):
    """
    Constrain a vector to be of a given base type.
    """

    expected_type: Attribute
    """The expected vector base type."""

    def verify(self, attr: Attribute, constraint_vars: dict[str, Attribute]) -> None:
        if not isinstance(attr, VectorType):
            raise VerifyException(f"{attr} should be of type VectorType.")
        attr = cast(VectorType[Attribute], attr)
        if attr.element_type != self.expected_type:
            raise VerifyException(
                f"Expected vector type to be {self.expected_type}, got {attr.element_type}."
            )


@dataclass(frozen=True)
class VectorBaseTypeAndRankConstraint(AttrConstraint):
    """
    Constrain a vector to be of a given rank and base type.
    """

    expected_type: Attribute
    """The expected vector base type."""

    expected_rank: int
    """The expected vector rank."""

    def verify(self, attr: Attribute, constraint_vars: dict[str, Attribute]) -> None:
        constraint = AllOf(
            (
                VectorBaseTypeConstraint(self.expected_type),
                VectorRankConstraint(self.expected_rank),
            )
        )
        constraint.verify(attr, constraint_vars)


@irdl_attr_definition
class DenseIntOrFPElementsAttr(
    ParametrizedAttribute, ContainerType[IntegerType | IndexType | AnyFloat]
):
    name = "dense"
    type: ParameterDef[
        RankedVectorOrTensorOf[IntegerType]
        | RankedVectorOrTensorOf[IndexType]
        | RankedVectorOrTensorOf[AnyFloat]
    ]
    data: ParameterDef[ArrayAttr[AnyIntegerAttr] | ArrayAttr[AnyFloatAttr]]

    # The type stores the shape data
    def get_shape(self) -> tuple[int, ...] | None:
        if isinstance(self.type, UnrankedTensorType):
            return None
        return self.type.get_shape()

    def get_element_type(self) -> IntegerType | IndexType | AnyFloat:
        return self.type.get_element_type()

    @property
    def shape_is_complete(self) -> bool:
        shape = self.get_shape()
        if shape is None or not len(shape):
            return False

        n = 1
        for dim in shape:
            if dim < 1:
                # Dimensions need to be greater or equal to one
                return False
            n *= dim

        # Product of dimensions needs to equal length
        return n == len(self.data.data)

    @staticmethod
    def create_dense_index(
        type: RankedVectorOrTensorOf[IndexType],
        data: Sequence[int] | Sequence[IntegerAttr[IndexType]],
    ) -> DenseIntOrFPElementsAttr:
        if len(data) and isinstance(data[0], int):
            attr_list = [
                IntegerAttr.from_index_int_value(d) for d in cast(Sequence[int], data)
            ]
        else:
            attr_list = cast(Sequence[IntegerAttr[IndexType]], data)

        return DenseIntOrFPElementsAttr([type, ArrayAttr(attr_list)])

    @staticmethod
    def create_dense_int(
        type: RankedVectorOrTensorOf[IntegerType],
        data: Sequence[int] | Sequence[IntegerAttr[IntegerType]],
    ) -> DenseIntOrFPElementsAttr:
        if len(data) and isinstance(data[0], int):
            attr_list = [
                IntegerAttr[IntegerType](d, type.element_type)
                for d in cast(Sequence[int], data)
            ]
        else:
            attr_list = cast(Sequence[IntegerAttr[IntegerType]], data)

        return DenseIntOrFPElementsAttr([type, ArrayAttr(attr_list)])

    @staticmethod
    def create_dense_float(
        type: RankedVectorOrTensorOf[AnyFloat],
        data: Sequence[int | float] | Sequence[AnyFloatAttr],
    ) -> DenseIntOrFPElementsAttr:
        if len(data) and isinstance(data[0], int | float):
            attr_list = [
                FloatAttr(float(d), type.element_type)
                for d in cast(Sequence[int | float], data)
            ]
        else:
            attr_list = cast(Sequence[AnyFloatAttr], data)

        return DenseIntOrFPElementsAttr([type, ArrayAttr(attr_list)])

    @overload
    @staticmethod
    def from_list(
        type: RankedVectorOrTensorOf[AnyFloat | IntegerType | IndexType],
        data: (
            Sequence[int]
            | Sequence[IntegerAttr[IndexType]]
            | Sequence[IntegerAttr[IntegerType]]
        ),
    ) -> DenseIntOrFPElementsAttr: ...

    @overload
    @staticmethod
    def from_list(
        type: RankedVectorOrTensorOf[AnyFloat | IntegerType | IndexType],
        data: Sequence[int | float] | Sequence[AnyFloatAttr],
    ) -> DenseIntOrFPElementsAttr: ...

    @staticmethod
    def from_list(
        type: RankedVectorOrTensorOf[AnyFloat | IntegerType | IndexType],
        data: Sequence[int | float] | Sequence[AnyIntegerAttr] | Sequence[AnyFloatAttr],
    ) -> DenseIntOrFPElementsAttr:
        if isinstance(type.element_type, AnyFloat):
            new_type = cast(RankedVectorOrTensorOf[AnyFloat], type)
            new_data = cast(Sequence[int | float] | Sequence[FloatAttr[AnyFloat]], data)
            return DenseIntOrFPElementsAttr.create_dense_float(new_type, new_data)

        match type.element_type:
            case IntegerType():
                new_type = cast(RankedVectorOrTensorOf[IntegerType], type)
                new_data = cast(
                    Sequence[int] | Sequence[IntegerAttr[IntegerType]], data
                )
                return DenseIntOrFPElementsAttr.create_dense_int(new_type, new_data)
            case IndexType():
                new_type = cast(RankedVectorOrTensorOf[IndexType], type)
                new_data = cast(Sequence[int] | Sequence[IntegerAttr[IndexType]], data)
                return DenseIntOrFPElementsAttr.create_dense_index(new_type, new_data)

    @staticmethod
    def vector_from_list(
        data: Sequence[int] | Sequence[float],
        data_type: IntegerType | IndexType | AnyFloat,
    ) -> DenseIntOrFPElementsAttr:
        t = VectorType(data_type, [len(data)])
        return DenseIntOrFPElementsAttr.from_list(t, data)

    @staticmethod
    def tensor_from_list(
        data: (
            Sequence[int]
            | Sequence[float]
            | Sequence[IntegerAttr[IndexType]]
            | Sequence[IntegerAttr[IntegerType]]
            | Sequence[AnyFloatAttr]
        ),
        data_type: IntegerType | IndexType | AnyFloat,
        shape: Sequence[int],
    ) -> DenseIntOrFPElementsAttr:
        t = TensorType(data_type, shape)
        return DenseIntOrFPElementsAttr.from_list(t, data)


@irdl_attr_definition
class DenseResourceAttr(ParametrizedAttribute):
    name = "dense_resource"

    resource_handle: ParameterDef[StringAttr]

    # Should be a ShapedType, but this is not defined yet in xDSL
    type: ParameterDef[Attribute]

    @staticmethod
    def from_params(handle: str | StringAttr, type: Attribute) -> DenseResourceAttr:
        if isinstance(handle, str):
            handle = StringAttr(handle)
        return DenseResourceAttr([handle, type])


@irdl_attr_definition
class DenseArrayBase(ParametrizedAttribute):
    name = "array"

    elt_type: ParameterDef[IntegerType | IndexType | AnyFloat]
    data: ParameterDef[ArrayAttr[IntAttr] | ArrayAttr[FloatData]]

    def verify(self):
        if isinstance(self.elt_type, IntegerType | IndexType):
            for d in self.data.data:
                if isinstance(d, FloatData):
                    raise VerifyException(
                        "dense array of integer or index element type "
                        "should only contain integers"
                    )
        else:
            for d in self.data.data:
                if isinstance(d, IntAttr):
                    raise VerifyException(
                        "dense array of float element type "
                        "should only contain floats"
                    )

    @staticmethod
    def create_dense_int_or_index(
        data_type: IntegerType | IndexType, data: Sequence[int] | Sequence[IntAttr]
    ) -> DenseArrayBase:
        if len(data) and isinstance(data[0], int):
            attr_list = [IntAttr(d) for d in cast(Sequence[int], data)]
        else:
            attr_list = cast(Sequence[IntAttr], data)

        return DenseArrayBase([data_type, ArrayAttr(attr_list)])

    @staticmethod
    def create_dense_float(
        data_type: AnyFloat, data: Sequence[int | float] | Sequence[FloatData]
    ) -> DenseArrayBase:
        if len(data) and isinstance(data[0], int | float):
            attr_list = [FloatData(float(d)) for d in cast(Sequence[int | float], data)]
        else:
            attr_list = cast(Sequence[FloatData], data)

        return DenseArrayBase([data_type, ArrayAttr(attr_list)])

    @overload
    @staticmethod
    def from_list(
        data_type: IntegerType | IndexType, data: Sequence[int] | Sequence[IntAttr]
    ) -> DenseArrayBase: ...

    @overload
    @staticmethod
    def from_list(
        data_type: Attribute, data: Sequence[int | float] | Sequence[FloatData]
    ) -> DenseArrayBase: ...

    @staticmethod
    def from_list(
        data_type: Attribute,
        data: (
            Sequence[int]
            | Sequence[int | float]
            | Sequence[IntAttr]
            | Sequence[FloatData]
        ),
    ) -> DenseArrayBase:
        if isinstance(data_type, IndexType | IntegerType):
            _data = cast(Sequence[int] | Sequence[IntAttr], data)
            return DenseArrayBase.create_dense_int_or_index(data_type, _data)
        elif isinstance(data_type, AnyFloat):
            _data = cast(Sequence[int | float] | Sequence[FloatData], data)
            return DenseArrayBase.create_dense_float(data_type, _data)
        else:
            raise TypeError(f"Unsupported element type {data_type}")

    def as_tuple(self) -> tuple[int, ...] | tuple[float, ...]:
        """
        Get the "raw" data out as a tuple. This will not
        apply the datatype restrictions that the array element
        type would suggest!

        e.g. given a dense<i8: 99999999, 255, 256>, as_tuple()
        would return 1234567, 255, 256 and not 135, 255, 0 (mod 256)
        """
        return tuple(x.data for x in self.data.data)


@irdl_attr_definition
class FunctionType(ParametrizedAttribute, TypeAttribute):
    name = "fun"

    inputs: ParameterDef[ArrayAttr[Attribute]]
    outputs: ParameterDef[ArrayAttr[Attribute]]

    @staticmethod
    def from_lists(
        inputs: Sequence[Attribute], outputs: Sequence[Attribute]
    ) -> FunctionType:
        return FunctionType([ArrayAttr(inputs), ArrayAttr(outputs)])

    @staticmethod
    def from_attrs(
        inputs: ArrayAttr[Attribute], outputs: ArrayAttr[Attribute]
    ) -> FunctionType:
        return FunctionType([inputs, outputs])


@irdl_attr_definition
class OpaqueAttr(ParametrizedAttribute):
    name = "opaque"

    ident: ParameterDef[StringAttr]
    value: ParameterDef[StringAttr]
    type: ParameterDef[Attribute]

    @staticmethod
    def from_strings(name: str, value: str, type: Attribute = NoneAttr()) -> OpaqueAttr:
        return OpaqueAttr([StringAttr(name), StringAttr(value), type])


@irdl_attr_definition
class StridedLayoutAttr(ParametrizedAttribute):
    """
    An attribute representing a strided layout of a shaped type.
    See https://mlir.llvm.org/docs/Dialects/Builtin/#stridedlayoutattr

    Contrary to MLIR, we represent dynamic offsets and strides with
    `NoneAttr`, and we do not restrict offsets and strides to 64-bits
    integers.
    """

    name = "strided"

    strides: ParameterDef[ArrayAttr[IntAttr | NoneAttr]]
    offset: ParameterDef[IntAttr | NoneAttr]

    def __init__(
        self,
        strides: (
            ArrayAttr[IntAttr | NoneAttr] | Sequence[int | None | IntAttr | NoneAttr]
        ),
        offset: int | None | IntAttr | NoneAttr = 0,
    ) -> None:
        if not isinstance(strides, ArrayAttr):
            strides_values: list[IntAttr | NoneAttr] = []
            for stride in strides:
                if isinstance(stride, int):
                    strides_values.append(IntAttr(stride))
                elif stride is None:
                    strides_values.append(NoneAttr())
                else:
                    strides_values.append(stride)
            strides = ArrayAttr(strides_values)

        if isinstance(offset, int):
            offset = IntAttr(offset)
        if offset is None:
            offset = NoneAttr()

        super().__init__([strides, offset])


@irdl_attr_definition
class AffineMapAttr(Data[AffineMap]):
    """An Attribute containing an AffineMap object."""

    name = "affine_map"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> AffineMap:
        with parser.in_angle_brackets():
            data = parser.parse_affine_map()
            return data

    def print_parameter(self, printer: Printer) -> None:
        printer.print_string(f"{self.data}")

    @staticmethod
    def constant_map(value: int) -> AffineMapAttr:
        return AffineMapAttr(AffineMap.constant_map(value))


@irdl_attr_definition
class AffineSetAttr(Data[AffineSet]):
    """An attribute containing an AffineSet object."""

    name = "affine_set"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> AffineSet:
        return parser.parse_affine_set()

    def print_parameter(self, printer: Printer) -> None:
        printer.print_string(f"{self.data}")


@irdl_op_definition
class UnrealizedConversionCastOp(IRDLOperation):
    name = "builtin.unrealized_conversion_cast"

    inputs: VarOperand = var_operand_def()
    outputs: VarOpResult = var_result_def()

    @staticmethod
    def get(inputs: Sequence[SSAValue | Operation], result_type: Sequence[Attribute]):
        return UnrealizedConversionCastOp.build(
            operands=[inputs],
            result_types=[result_type],
        )

    @classmethod
    def parse(cls, parser: Parser) -> Self:
        if parser.parse_optional_characters("to") is None:
            args = parser.parse_comma_separated_list(
                parser.Delimiter.NONE,
                parser.parse_unresolved_operand,
            )
            parser.parse_punctuation(":")
            input_types = parser.parse_comma_separated_list(
                parser.Delimiter.NONE,
                parser.parse_type,
            )
            parser.parse_characters("to")
            inputs = parser.resolve_operands(args, input_types, parser.pos)
        else:
            inputs = list[SSAValue]()
        output_types = parser.parse_comma_separated_list(
            parser.Delimiter.NONE,
            parser.parse_type,
        )
        attributes = parser.parse_optional_attr_dict()
        return cls(
            operands=[inputs], result_types=[output_types], attributes=attributes
        )

    def print(self, printer: Printer):
        def print_fn(operand: SSAValue) -> None:
            return printer.print_attribute(operand.type)

        if self.inputs:
            printer.print(" ")
            printer.print_list(self.inputs, printer.print_operand)
            printer.print_string(" : ")
            printer.print_list(self.inputs, print_fn)
        printer.print_string(" to ")
        printer.print_list(self.outputs, print_fn)
        printer.print_op_attributes(self.attributes)


class UnregisteredOp(Operation, ABC):
    """
    An unregistered operation.

    Each unregistered op is registered as a subclass of `UnregisteredOp`,
    and op with different names have distinct subclasses.
    """

    name = "builtin.unregistered"
    traits = frozenset()

    @property
    def op_name(self) -> StringAttr:
        if "op_name__" not in self.attributes:
            raise ValueError("missing 'op_name__' attribute")
        op_name = self.attributes["op_name__"]
        if not isinstance(op_name, StringAttr):
            raise ValueError(
                f"'op_name__' is expected to have 'StringAttr' type, got {op_name}"
            )
        return op_name

    @classmethod
    def with_name(cls, name: str) -> type[Operation]:
        """
        Return a new unregistered operation type given a name.
        This function should not be called directly. Use methods from
        `MLContext` to get an `UnregisteredOp` type.
        """

        class UnregisteredOpWithName(UnregisteredOp):
            @classmethod
            def create(
                cls,
                *,
                operands: Sequence[SSAValue] = (),
                result_types: Sequence[Attribute] = (),
                properties: Mapping[str, Attribute] = {},
                attributes: Mapping[str, Attribute] = {},
                successors: Sequence[Block] = (),
                regions: Sequence[Region] = (),
            ):
                op = super().create(
                    operands=operands,
                    result_types=result_types,
                    properties=properties,
                    attributes=attributes,
                    successors=successors,
                    regions=regions,
                )
                op.attributes["op_name__"] = StringAttr(name)
                return op

        return UnregisteredOpWithName


class UnregisteredAttr(ParametrizedAttribute, ABC):
    """
    An unregistered attribute or type.

    Each unregistered attribute is registered as a subclass of
    `UnregisteredAttr`, and attribute with different names have
    distinct subclasses.

    Since attributes do not have a generic format, unregistered
    attributes represent their original parameters as a string,
    which is exactly the content parsed from the textual
    representation.
    """

    name = "builtin.unregistered"

    attr_name: ParameterDef[StringAttr]
    is_type: ParameterDef[IntAttr]
    is_opaque: ParameterDef[IntAttr]
    value: ParameterDef[StringAttr]
    """
    This parameter is non-null is the attribute is a type, and null otherwise.
    """

    def __init__(
        self,
        attr_name: str | StringAttr,
        is_type: bool | IntAttr,
        is_opaque: bool | IntAttr,
        value: str | StringAttr,
    ):
        if isinstance(attr_name, str):
            attr_name = StringAttr(attr_name)
        if isinstance(is_type, bool):
            is_type = IntAttr(int(is_type))
        if isinstance(is_opaque, bool):
            is_opaque = IntAttr(int(is_opaque))
        if isinstance(value, str):
            value = StringAttr(value)
        super().__init__([attr_name, is_type, is_opaque, value])

    @classmethod
    def with_name_and_type(cls, name: str, is_type: bool) -> type[UnregisteredAttr]:
        """
        Return a new unregistered attribute type given a name and a
        boolean indicating if the attribute can be a type.
        This function should not be called directly. Use methods from
        `MLContext` to get an `UnregisteredAttr` type.
        """

        @irdl_attr_definition
        class UnregisteredAttrWithName(UnregisteredAttr):
            def verify(self):
                if self.attr_name.data != name:
                    raise VerifyException("Unregistered attribute name mismatch")
                if self.is_type.data != int(is_type):
                    raise VerifyException("Unregistered attribute is_type mismatch")

        @irdl_attr_definition
        class UnregisteredAttrTypeWithName(UnregisteredAttr, TypeAttribute):
            def verify(self):
                if self.attr_name.data != name:
                    raise VerifyException("Unregistered attribute name mismatch")
                if self.is_type.data != int(is_type):
                    raise VerifyException("Unregistered attribute is_type mismatch")

        if is_type:
            return UnregisteredAttrWithName
        else:
            return UnregisteredAttrTypeWithName


@irdl_op_definition
class ModuleOp(IRDLOperation):
    name = "builtin.module"

    sym_name = opt_attr_def(StringAttr)

    body: Region = region_def("single_block")

    traits = frozenset(
        [
            IsolatedFromAbove(),
            NoTerminator(),
            OptionalSymbolOpInterface(),
            SymbolTable(),
        ]
    )

    def __init__(
        self,
        ops: list[Operation] | Region,
        attributes: Mapping[str, Attribute] | None = None,
    ):
        if attributes is None:
            attributes = {}
        if isinstance(ops, Region):
            region = ops
        else:
            region = Region(Block(ops))
        super().__init__(regions=[region], attributes=attributes)

    @property
    def ops(self) -> BlockOps:
        return self.body.ops

    @classmethod
    def parse(cls, parser: Parser) -> ModuleOp:
        attributes = parser.parse_optional_attr_dict_with_keyword()
        if attributes is not None:
            attributes = attributes.data
        region = parser.parse_region()

        # Add a block if the region is empty
        if len(region.blocks) == 0:
            region.add_block(Block())

        return ModuleOp(region, attributes)

    def print(self, printer: Printer) -> None:
        if len(self.attributes) != 0:
            printer.print(" attributes ")
            printer.print_op_attributes(self.attributes)

        if not self.body.block.ops:
            # Do not print the entry block if the region has an empty block
            printer.print(" {\n")
            printer.print("}")
        else:
            printer.print(" ", self.body)


# FloatXXType shortcuts
bf16 = BFloat16Type()
f16 = Float16Type()
f32 = Float32Type()
f64 = Float64Type()
f80 = Float64Type()
f128 = Float64Type()


_MemRefTypeElement = TypeVar("_MemRefTypeElement", bound=Attribute)
_UnrankedMemrefTypeElems = TypeVar(
    "_UnrankedMemrefTypeElems", bound=Attribute, covariant=True
)
_UnrankedMemrefTypeElemsInit = TypeVar("_UnrankedMemrefTypeElemsInit", bound=Attribute)


@irdl_attr_definition
class NoneType(ParametrizedAttribute, TypeAttribute):
    name = "none_type"


@irdl_attr_definition
class MemRefType(
    Generic[_MemRefTypeElement],
    ParametrizedAttribute,
    TypeAttribute,
    ShapedType,
    ContainerType[_MemRefTypeElement],
):
    name = "memref"

    shape: ParameterDef[ArrayAttr[IntAttr]]
    element_type: ParameterDef[_MemRefTypeElement]
    layout: ParameterDef[Attribute]
    memory_space: ParameterDef[Attribute]

    def __init__(
        self: MemRefType[_MemRefTypeElement],
        element_type: _MemRefTypeElement,
        shape: Iterable[int | IntAttr],
        layout: Attribute = NoneAttr(),
        memory_space: Attribute = NoneAttr(),
    ):
        shape = ArrayAttr(
            [IntAttr(dim) if isinstance(dim, int) else dim for dim in shape]
        )
        super().__init__(
            [
                shape,
                element_type,
                layout,
                memory_space,
            ]
        )

    def get_num_dims(self) -> int:
        return len(self.shape.data)

    def get_shape(self) -> tuple[int, ...]:
        return tuple(i.data for i in self.shape.data)

    def get_element_type(self) -> _MemRefTypeElement:
        return self.element_type

    @deprecated_constructor
    @staticmethod
    def from_element_type_and_shape(
        referenced_type: _MemRefTypeElement,
        shape: Iterable[int | AnyIntegerAttr],
        layout: Attribute = NoneAttr(),
        memory_space: Attribute = NoneAttr(),
    ) -> MemRefType[_MemRefTypeElement]:
        shape_int = [i if isinstance(i, int) else i.value.data for i in shape]
        return MemRefType(referenced_type, shape_int, layout, memory_space)

    @deprecated_constructor
    @staticmethod
    def from_params(
        referenced_type: _MemRefTypeElement,
        shape: ArrayAttr[AnyIntegerAttr] = ArrayAttr(
            [IntegerAttr.from_int_and_width(1, 64)]
        ),
        layout: Attribute = NoneAttr(),
        memory_space: Attribute = NoneAttr(),
    ) -> MemRefType[_MemRefTypeElement]:
        shape_int = [i.value.data for i in shape.data]
        return MemRefType(referenced_type, shape_int, layout, memory_space)

    @classmethod
    def parse_parameters(cls, parser: AttrParser) -> list[Attribute]:
        parser.parse_punctuation("<", " in memref attribute")
        shape = parser.parse_attribute()
        parser.parse_punctuation(",", " between shape and element type parameters")
        type = parser.parse_attribute()
        # If we have a layout or a memory space, parse both of them.
        if parser.parse_optional_punctuation(",") is None:
            parser.parse_punctuation(">", " at end of memref attribute")
            return [shape, type, NoneAttr(), NoneAttr()]
        layout = parser.parse_attribute()
        parser.parse_punctuation(",", " between layout and memory space")
        memory_space = parser.parse_attribute()
        parser.parse_punctuation(">", " at end of memref attribute")

        return [shape, type, layout, memory_space]

    def print_parameters(self, printer: Printer) -> None:
        printer.print("<", self.shape, ", ", self.element_type)
        if self.layout != NoneAttr() or self.memory_space != NoneAttr():
            printer.print(", ", self.layout, ", ", self.memory_space)
        printer.print(">")


@irdl_attr_definition
class UnrankedMemrefType(
    Generic[_UnrankedMemrefTypeElems], ParametrizedAttribute, TypeAttribute
):
    name = "unranked_memref"

    element_type: ParameterDef[_UnrankedMemrefTypeElems]
    memory_space: ParameterDef[Attribute]

    @staticmethod
    def from_type(
        referenced_type: _UnrankedMemrefTypeElemsInit,
        memory_space: Attribute = NoneAttr(),
    ) -> UnrankedMemrefType[_UnrankedMemrefTypeElemsInit]:
        return UnrankedMemrefType([referenced_type, memory_space])


AnyUnrankedMemrefType: TypeAlias = UnrankedMemrefType[Attribute]

Builtin = Dialect(
    "builtin",
    [
        ModuleOp,
        UnregisteredOp,
        UnrealizedConversionCastOp,
    ],
    [
        UnregisteredAttr,
        # Attributes
        StringAttr,
        SymbolRefAttr,
        SymbolNameAttr,
        IntAttr,
        IntegerAttr,
        ArrayAttr,
        DictionaryAttr,
        DenseIntOrFPElementsAttr,
        DenseResourceAttr,
        UnitAttr,
        FloatData,
        LocationAttr,
        NoneAttr,
        OpaqueAttr,
        # Types
        ComplexType,
        FunctionType,
        BFloat16Type,
        Float16Type,
        Float32Type,
        Float64Type,
        Float80Type,
        Float128Type,
        FloatAttr,
        SignednessAttr,
        TupleType,
        IntegerType,
        IndexType,
        NoneType,
        VectorType,
        TensorType,
        UnrankedTensorType,
        AffineMapAttr,
        AffineSetAttr,
        MemRefType,
        UnrankedMemrefType,
    ],
)



from dataclasses import dataclass

from xdsl.backend.riscv.lowering.convert_riscv_scf_to_riscv_cf import (
    ConvertRiscvScfToRiscvCfPass,
)
from xdsl.backend.riscv.lowering.convert_snitch_stream_to_snitch import (
    ConvertSnitchStreamToSnitch,
)
from xdsl.passes import ModulePass, PipelinePass
from xdsl.transforms.canonicalize import CanonicalizePass
from xdsl.transforms.convert_riscv_scf_for_to_frep import ConvertRiscvScfForToFrepPass
from xdsl.transforms.lower_snitch import LowerSnitchPass
from xdsl.transforms.riscv_register_allocation import RISCVRegisterAllocation
from xdsl.transforms.riscv_scf_loop_range_folding import RiscvScfLoopRangeFoldingPass
from xdsl.transforms.snitch_register_allocation import SnitchRegisterAllocation

TEST_LOWER_SNITCH_STREAM_TO_ASM_PASSES: tuple[ModulePass, ...] = (
    ConvertRiscvScfForToFrepPass(),
    SnitchRegisterAllocation(),
    ConvertSnitchStreamToSnitch(),
    LowerSnitchPass(),
    CanonicalizePass(),
    RiscvScfLoopRangeFoldingPass(),
    CanonicalizePass(),
    RISCVRegisterAllocation(),
    CanonicalizePass(),
    ConvertRiscvScfToRiscvCfPass(),
    CanonicalizePass(),
)


@dataclass(frozen=True)
class TestLowerSnitchStreamToAsm(PipelinePass):
    """
    A compiler pass used for testing of the lowering from ML kernels defined as
    snitch_stream + riscv operations to riscv-assemby leveraging Snitch cores.
    """

    name = "test-lower-snitch-stream-to-asm"

    passes: tuple[ModulePass, ...] = TEST_LOWER_SNITCH_STREAM_TO_ASM_PASSES



from collections.abc import Sequence
from io import StringIO
from typing import Any, cast

import wgpu  # pyright: ignore
import wgpu.utils  # pyright: ignore

from xdsl.dialects import gpu
from xdsl.dialects.builtin import IndexType
from xdsl.dialects.memref import MemRefType
from xdsl.interpreter import Interpreter, InterpreterFunctions, impl, register_impls
from xdsl.interpreters.experimental.wgsl_printer import WGSLPrinter
from xdsl.interpreters.shaped_array import ShapedArray
from xdsl.ir import Attribute, SSAValue
from xdsl.traits import SymbolTable
from xdsl.utils.hints import isa


@register_impls
class WGPUFunctions(InterpreterFunctions):
    device = cast(wgpu.GPUDevice, wgpu.utils.get_default_device())
    shader_modules: dict[gpu.FuncOp, wgpu.GPUShaderModule] = {}

    def buffer_from_operand(self, interpreter: Interpreter, operand: SSAValue):
        """
        Prepare a GPUBuffer from an SSA operand.

        memrefs are excpected to be GPUBuffers at this point.

        Still a helper function, because boilerplating will need to happen to forward
        e.g. scalar parameters!
        """
        if isa(operand.type, MemRefType[Attribute]):
            value = interpreter.get_values((operand,))[0]
            if not isinstance(value, wgpu.GPUBuffer):
                raise ValueError(
                    "gpu.launch_func memref operand expected to be GPU-allocated"
                )
            return value
        raise NotImplementedError(f"{operand.type} not yet mapped to WGPU.")

    def prepare_bindings(
        self, interpreter: Interpreter, kernel_operands: Sequence[SSAValue]
    ):
        """
        Boilerplate preparation for arguments bindings.
        """
        layouts: list[dict[str, Any]] = []
        bindings: list[dict[str, Any]] = []
        for i, o in enumerate(kernel_operands):
            buffer = WGPUFunctions.buffer_from_operand(self, interpreter, o)

            layouts.append(
                {
                    "binding": i,
                    "visibility": wgpu.ShaderStage.COMPUTE,  # pyright: ignore
                    "buffer": {
                        "type": wgpu.BufferBindingType.storage  # pyright: ignore
                    },
                }
            )
            bindings.append(
                {
                    "binding": i,
                    "resource": {"buffer": buffer, "offset": 0, "size": buffer.size},
                }
            )

        return layouts, bindings

    def compile_func(self, op: gpu.FuncOp):
        """
        Compile a gpu.func if not already done.
        """
        if op not in self.shader_modules:
            wgsl_printer = WGSLPrinter()
            wgsl_source = StringIO("")
            wgsl_printer.print(op, wgsl_source)
            self.shader_modules[op] = cast(
                wgpu.GPUShaderModule,
                self.device.create_shader_module(  # pyright: ignore
                    code=wgsl_source.getvalue()
                ),  # pyright: ignore
            )

    @impl(gpu.AllocOp)
    def run_alloc(
        self, interpreter: Interpreter, op: gpu.AllocOp, args: tuple[Any, ...]
    ):
        """
        Allocate a GPUBuffer according to a gpu.alloc operation, return it as the memref
        value.
        """
        if args or op.asyncToken:
            raise NotImplementedError(
                "Only synchronous, known-sized gpu.alloc implemented yet."
            )
        memref_type = cast(MemRefType[Attribute], op.result.type)
        match (memref_type.element_type):
            case IndexType():
                element_size = 4
            case _:
                raise NotImplementedError(
                    f"The element type {memref_type.element_type} for gpu.alloc is not implemented yet."
                )
        buffer = cast(
            wgpu.GPUBuffer,
            self.device.create_buffer(  # pyright: ignore
                size=memref_type.element_count() * element_size,
                usage=wgpu.BufferUsage.STORAGE  # pyright: ignore
                | wgpu.BufferUsage.COPY_SRC,  # pyright: ignore
            ),
        )
        return (buffer,)

    @impl(gpu.MemcpyOp)
    def run_memcpy(
        self, interpreter: Interpreter, op: gpu.MemcpyOp, args: tuple[Any, ...]
    ) -> tuple[()]:
        """
        Copy buffers according to the gpu.memcpy operation.

        Only Device to Host copy is implemented here, to keep the first draft bearable.
        """
        src, dst = interpreter.get_values((op.src, op.dst))
        if not (isinstance(src, wgpu.GPUBuffer) and isinstance(dst, ShapedArray)):
            raise NotImplementedError(
                f"Only device to host copy is implemented for now. got {src} to {dst}"
            )

        # Get device/source view
        memview = cast(
            memoryview, self.device.queue.read_buffer(src)  # pyright: ignore
        )
        dst_type = cast(MemRefType[Attribute], op.dst.type)
        match (dst_type.element_type):
            case IndexType():
                format = "I"
            case _:
                raise NotImplementedError(
                    f"copy for element type {dst_type.element_type} not yet implemented."
                )
        memview = memview.cast(format, dst_type.get_shape())
        for index in dst.indices():
            dst.store(index, memview.__getitem__(index))  # pyright: ignore
        return ()

    @impl(gpu.LaunchFuncOp)
    def run_launch_func(
        self, interpreter: Interpreter, op: gpu.LaunchFuncOp, args: tuple[Any, ...]
    ):
        """
        Launch a GPU kernel through the WebGPU API.
        """
        if op.asyncToken is not None or op.asyncDependencies:
            raise NotImplementedError(
                "The WGPU interpreter does not handle asynchronous GPU regions at the moment."
            )

        gridSize = interpreter.get_values((op.gridSizeX, op.gridSizeY, op.gridSizeZ))
        blockSize = interpreter.get_values(
            (op.blockSizeX, op.blockSizeY, op.blockSizeZ)
        )
        kernel_operands = op.kernelOperands

        func = SymbolTable.lookup_symbol(op, op.kernel)
        assert isinstance(func, gpu.FuncOp)
        WGPUFunctions.compile_func(self, func)
        shader_module = self.shader_modules[func]

        # Compute the dispatch number
        # If the func has a known block size, it's reflected in the compiled module
        # Otherwise, it defaults to (1,1,1) currently and we have to take this
        # into account
        if func.known_block_size:
            dispatch = gridSize
        else:
            dispatch = [a * b for a, b in zip(gridSize, blockSize)]

        layouts, bindings = WGPUFunctions.prepare_bindings(
            self, interpreter, kernel_operands
        )

        # All the boilerplate
        device = self.device
        # Put bindings together
        bind_group_layout = device.create_bind_group_layout(  # pyright: ignore
            entries=layouts
        )
        pipeline_layout = device.create_pipeline_layout(  # pyright: ignore
            bind_group_layouts=[bind_group_layout]
        )
        bind_group = device.create_bind_group(  # pyright: ignore
            layout=bind_group_layout, entries=bindings  # pyright: ignore
        )

        # Create and run the pipeline
        compute_pipeline = device.create_compute_pipeline(  # pyright: ignore
            layout=pipeline_layout,  # pyright: ignore
            compute={"module": shader_module, "entry_point": func.sym_name.data},
        )

        command_encoder = device.create_command_encoder()  # pyright: ignore
        compute_pass = command_encoder.begin_compute_pass()  # pyright: ignore
        compute_pass.set_pipeline(compute_pipeline)  # pyright: ignore
        compute_pass.set_bind_group(  # pyright: ignore
            0, bind_group, [], 0, 0
        )  # last 2 elements not used
        compute_pass.dispatch_workgroups(*dispatch)  # x y z # pyright: ignore
        compute_pass.end()  # pyright: ignore
        device.queue.submit([command_encoder.finish()])  # pyright: ignore

        # gpu.launch_func has no return
        return ()



import pytest

from xdsl.dialects.builtin import StringAttr, i32
from xdsl.ir import Block, BlockArgument, OpResult, SSAValue
from xdsl.irdl import IRDLOperation, irdl_op_definition, result_def


@irdl_op_definition
class TwoResultOp(IRDLOperation):
    name = "test.tworesults"

    res1: OpResult = result_def(StringAttr)
    res2: OpResult = result_def(StringAttr)


def test_var_mixed_builder():
    op = TwoResultOp.build(result_types=[StringAttr("0"), StringAttr("2")])

    with pytest.raises(ValueError):
        _ = SSAValue.get(op)


@pytest.mark.parametrize(
    "name",
    [
        "test",
        "-2",
        "test123",
        "kebab-case-name",
        None,
    ],
)
def test_ssa_value_name_hints(name: str | None):
    r"""
    As per the MLIR language reference, legal SSA value names must conform to
        ([0-9]+|([A-Za-z_$.-][\w$.-]*))

    https://mlir.llvm.org/docs/LangRef/#identifiers-and-keywords

    xDSL SSA value name hints are a refinement of these rules.
    We only accept non-numeric name hints, because the printer will
    generate its own numeric names.

    This test tests valid name hints:
    """
    val = BlockArgument(i32, Block(), 0)

    val.name_hint = name
    assert val.name_hint == name


@pytest.mark.parametrize("name", ["&", "#", "%2", '"', "::", "42"])
def test_invalid_ssa_vals(name: str):
    """
    This test tests invalid name hints that raise an error, because
    they don't conform to the rules of how SSA value names should be
    structured.
    """
    val = BlockArgument(i32, Block(), 0)
    with pytest.raises(ValueError):
        val.name_hint = name



#!/usr/bin/env python3

import argparse
import sys
from collections.abc import Sequence

from xdsl.interpreter import Interpreter
from xdsl.interpreters import (
    register_implementations,
)
from xdsl.ir import MLContext
from xdsl.parser import Parser
from xdsl.tools.command_line_tool import CommandLineTool
from xdsl.traits import CallableOpInterface


class xDSLRunMain(CommandLineTool):
    interpreter: Interpreter

    def __init__(
        self,
        description: str = "xDSL modular runner",
        args: Sequence[str] | None = None,
    ):
        self.available_frontends = {}

        self.ctx = MLContext()
        self.register_all_dialects()
        self.register_all_frontends()
        # arg handling
        arg_parser = argparse.ArgumentParser(description=description)
        self.register_all_arguments(arg_parser)
        self.args = arg_parser.parse_args(args=args)

        self.ctx.allow_unregistered = self.args.allow_unregistered_dialect

    def register_all_arguments(self, arg_parser: argparse.ArgumentParser):
        arg_parser.add_argument(
            "--wgpu",
            default=False,
            action="store_true",
            help="Enable the WGPU JIT-compilation interpreter.",
        )
        arg_parser.add_argument(
            "--onnx",
            default=False,
            action="store_true",
            help="Enable the onnx-compilation interpreter.",
        )
        arg_parser.add_argument(
            "--verbose",
            default=False,
            action="store_true",
            help="Print resulting Python values.",
        )
        arg_parser.add_argument(
            "--symbol",
            default="main",
            type=str,
            help="Name of function to call.",
        )
        arg_parser.add_argument(
            "--index-bitwidth",
            choices=(32, 64),
            type=int,
            nargs="?",
            help="Bitwidth of the index type representation.",
        )
        arg_parser.add_argument(
            "--args",
            default="",
            type=str,
            help="Arguments to pass to entry function. Comma-separated list of xDSL Attributes, that will be parsed and converted by the interpreter.",
        )
        return super().register_all_arguments(arg_parser)

    def register_implementations(self, interpreter: Interpreter):
        register_implementations(
            interpreter,
            self.ctx,
            include_wgpu=self.args.wgpu,
            include_onnx=self.args.onnx,
        )

    def run(self):
        input, file_extension = self.get_input_stream()
        try:
            module = self.parse_chunk(input, file_extension)
            if module is not None:
                module.verify()
                interpreter = Interpreter(
                    module, index_bitwidth=self.args.index_bitwidth
                )
                self.register_implementations(interpreter)
                symbol = self.args.symbol
                assert isinstance(symbol, str)
                parser = Parser(self.ctx, self.args.args, "args")
                runner_args = parser.parse_optional_undelimited_comma_separated_list(
                    parser.parse_optional_attribute, parser.parse_attribute
                )
                if runner_args is None:
                    runner_args = ()
                op = interpreter.get_op_for_symbol(symbol)
                trait = op.get_trait(CallableOpInterface)
                assert trait is not None

                args = tuple(
                    interpreter.value_for_attribute(attr, attr_type)
                    for attr, attr_type in zip(
                        runner_args, trait.get_argument_types(op)
                    )
                )
                result = interpreter.call_op(op, args)
                if self.args.verbose:
                    if result:
                        if len(result) == 1:
                            print(f"result: {result[0]}")
                        else:
                            print("result: (")
                            print(",\n".join(f"    {res}" for res in result))
                            print(")")
                    else:
                        print("result: ()")
        finally:
            if input is not sys.stdin:
                input.close()


def main():
    return xDSLRunMain().run()


if __name__ == "__main__":
    main()



from dataclasses import dataclass

from xdsl.dialects import builtin, memref, scf
from xdsl.ir import Block, Operation, Region, SSAValue
from xdsl.irdl import Operand
from xdsl.passes import MLContext, ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint


def find_same_target_store(load: memref.Load):
    """
    Find the corresponding store operation (same memeref target) for a load when there
    is only a single one within a block.
    """

    parent_block = load.parent_block()

    if parent_block is None:
        return None

    found_op = None

    for op in parent_block.ops:
        if (
            isinstance(op, memref.Store)
            and op.memref == load.memref
            and op.indices == load.indices
        ):
            if found_op is None:
                found_op = op
            else:
                return None

    return found_op


def is_loop_dependent(val: SSAValue, loop: scf.For):
    """
    Returns true if the SSA value is dependent by the induction varialbe of the loop.

    This is achieved by traversing the SSA use-def chain of the SSA value; if the
    induction variable contributes to the value, then it depends on it.
    """
    worklist: set[SSAValue] = set()
    visited: set[SSAValue] = set()

    worklist.add(val)

    while worklist:
        val = worklist.pop()
        visited.add(val)

        if val is loop.body.block.args[0]:
            return True

        if isinstance(val.owner, Operation):
            for oprnd in val.owner.operands:
                if oprnd not in visited:
                    worklist.add(oprnd)

    return False


@dataclass
class LoopHoistMemref(RewritePattern):
    """
    Hoist pairs of memref.loads and memref.stores out of a loop.

    This rewrite hoists pairs of memref.load and memref.store operations outside
    of their enclosing scf.loop. The memref operation pair is considered for this rewrite
    if their memref target location is the same and it is constant w.r.t. the induction
    variable of the containing loop.

    The functionality is intentionally very restricted and does not handle:

    - Multiple loads from the same location
    - Multiple stores of the same value and/or to same location
    - Loads must precede stores (i.e., WAR-only dependence)
    - Does not handle only loads or only stores
    - There is no consideration of aliasing (restricted only to same location memrefs)
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self,
        for_op: scf.For,
        rewriter: PatternRewriter,
    ) -> None:
        if for_op.parent_block() is None:
            return

        parent_block = for_op.body.block

        loads = [op for op in parent_block.ops if isinstance(op, memref.Load)]

        if not loads:
            return

        # filter out multiple loads from the same location
        load_locs = [load.memref for load in loads]
        dup_load_locs = [loc for loc in set(load_locs) if load_locs.count(loc) > 1]
        loads = [load for load in loads if load.memref not in dup_load_locs]

        load_store_pairs: dict[memref.Load, memref.Store] = {}

        for load in loads:
            if (
                (store := find_same_target_store(load))
                and parent_block.get_operation_index(load)
                < parent_block.get_operation_index(store)
                and not any(is_loop_dependent(idx, for_op) for idx in load.indices)
            ):
                load_store_pairs[load] = store

        # filter out stores using the same value
        store_vals = [store.value for store in load_store_pairs.values()]
        dup_store_vals = [val for val in store_vals if store_vals.count(val) > 1]
        load_store_pairs = {
            load: store
            for load, store in load_store_pairs.items()
            if store.value not in dup_store_vals
        }

        if not load_store_pairs:
            return

        # hoist new loads before the current loop
        new_loads = [load.clone() for load in load_store_pairs.keys()]
        rewriter.insert_op(new_loads, InsertPoint.before(for_op))

        new_body = Region()
        block_map: dict[Block, Block] = {}
        for_op.body.clone_into(new_body, None, None, block_map)

        load_map = {
            load: new_load
            for load, new_load in zip(for_op.body.block.ops, new_body.block.ops)
            if isinstance(load, memref.Load) and isinstance(new_load, memref.Load)
        }
        store_map = {
            store: new_store
            for store, new_store in zip(for_op.body.block.ops, new_body.block.ops)
            if isinstance(store, memref.Store) and isinstance(new_store, memref.Store)
        }

        new_block_args = [
            new_body.block.insert_arg(new_load.res.type, len(new_body.block.args))
            for new_load in new_loads
        ]

        toerase_ops: list[Operation] = []
        for new_block_arg, load in zip(new_block_args, load_store_pairs.keys()):
            interim_load = load_map[load]
            interim_load.res.replace_by(new_block_arg)
            toerase_ops.append(interim_load)

        new_yield_vals: list[Operand] = []
        for store in load_store_pairs.values():
            interim_store = store_map[store]
            new_yield_vals.append(interim_store.value)
            toerase_ops.append(interim_store)

        for op in toerase_ops:
            op.detach()
            op.erase()

        # yield the value that was used in the old store
        assert new_body.block.last_op is not None
        rewriter.replace_op(new_body.block.last_op, scf.Yield(*new_yield_vals))

        new_for_op = scf.For(for_op.lb, for_op.ub, for_op.step, new_loads, new_body)

        # use yielded results of new loop in stores after the loop
        new_stores = [
            memref.Store.get(new_for_op.res[idx], store.memref, store.indices)
            for idx, store in enumerate(load_store_pairs.values())
        ]
        rewriter.insert_op(new_stores, InsertPoint.after(for_op))

        rewriter.insert_op(new_for_op, InsertPoint.before(for_op))
        rewriter.erase_op(for_op)


@dataclass(frozen=True)
class LoopHoistMemrefPass(ModulePass):
    name = "loop-hoist-memref"

    def apply(self, ctx: MLContext, op: builtin.ModuleOp) -> None:
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    LoopHoistMemref(),
                ]
            ),
            walk_regions_first=True,
            apply_recursively=True,
        ).rewrite_module(op)



from __future__ import annotations

import math
from abc import ABC
from typing import Annotated, cast

from typing_extensions import Self

from xdsl.dialects.builtin import (
    Any,
    AnyFloat,
    AnyIntegerAttr,
    AnyTensorType,
    ArrayAttr,
    DenseIntOrFPElementsAttr,
    Float32Type,
    FloatAttr,
    IntegerAttr,
    IntegerType,
    MemRefType,
    NoneType,
    SSAValue,
    StringAttr,
    SymbolRefAttr,
    TensorType,
)
from xdsl.ir import (
    Attribute,
    Dialect,
)
from xdsl.irdl import (
    ConstraintVar,
    IRDLOperation,
    attr_def,
    irdl_op_definition,
    operand_def,
    opt_attr_def,
    result_def,
)
from xdsl.parser import Parser
from xdsl.printer import Printer
from xdsl.utils.exceptions import VerifyException


def verify_unidirectional_broadcast_shape(
    lhs: TensorType[Attribute], rhs: TensorType[Attribute], res: TensorType[Attribute]
) -> None:
    """
    Returns a unidirectional broadcastable shape
    """
    lhs_shape = lhs.get_shape()
    rhs_shape = rhs.get_shape()
    expected_shape = unidirectional_broadcast_shape(list(lhs_shape), list(rhs_shape))
    if expected_shape is None:
        raise VerifyException(
            f"operands have incompatible shapes: {lhs_shape} and {rhs_shape}"
        )
    res_type_shape = res.get_shape()
    if (
        len(expected_shape) != len(res_type_shape)
        or tuple(expected_shape) != res_type_shape
    ):
        raise VerifyException(
            f"result shape {expected_shape} does not match result type {res}"
        )


def verify_multidirectional_broadcast_shape(
    lhs: TensorType[Attribute], rhs: TensorType[Attribute], res: TensorType[Attribute]
) -> None:
    """
    Returns a multidirectional broadcastable shape
    """
    lhs_shape = lhs.get_shape()
    rhs_shape = rhs.get_shape()
    expected_shape = multidirectional_broadcast_shape(list(lhs_shape), list(rhs_shape))
    if expected_shape is None:
        raise VerifyException(
            f"operands have incompatible shapes: {lhs_shape} and {rhs_shape}"
        )
    res_type_shape = res.get_shape()
    if (
        len(expected_shape) != len(res_type_shape)
        or tuple(expected_shape) != res_type_shape
    ):
        raise VerifyException(
            f"result shape {expected_shape} does not match result type {res}"
        )


def unidirectional_broadcast_shape(lhs: list[int], rhs: list[int]) -> list[int] | None:
    """
    In ONNX, tensor B is unidirectional broadcastable to tensor A if one of the following is true:

    1. Tensor A and B both have exactly the same shape.
    2. Tensor A and B all have the same number of dimensions and
    the length of each dimensions is either a common length or B's length is 1.

    3.Tensor B has too few dimensions, and B can have its shapes prepended with a dimension of length 1 to satisfy
    property 2.

    When unidirectional broadcasting happens, the output's shape is the same as the shape of A (i.e.,
    the larger shape of two input tensors)

    https://github.com/onnx/onnx/blob/main/docs/Broadcasting.md
    """
    # Check if Tensor A and B both have exactly the same shape
    if lhs == rhs:
        return lhs

    lhs_len = len(lhs)
    rhs_len = len(rhs)
    prefix_len = lhs_len - rhs_len
    if prefix_len < 0:
        # lhs must not be shorter than rhs
        return None
    res_shape = lhs[:prefix_len]
    for dl, dr in zip(lhs[prefix_len:], rhs):
        if dl == dr or dr == 1 or dl == 1:
            res_shape.append(max(dl, dr))
        else:
            return None
    return res_shape


def multidirectional_broadcast_shape(
    lhs: list[int], rhs: list[int]
) -> list[int] | None:
    """
    In ONNX, a set of tensors are multidirectional broadcastable to the same shape if one of the following is true:

    1.The tensors all have exactly the same shape.
    2.The tensors all have the same number of dimensions and the length of each dimensions is either a common length or 1.
    3.The tensors that have too few dimensions can have their shapes prepended with a dimension of length 1 to satisfy property 2.

    https://github.com/onnx/onnx/blob/main/docs/Broadcasting.md
    """

    if len(lhs) > len(rhs):
        return unidirectional_broadcast_shape(rhs, lhs)
    else:
        return unidirectional_broadcast_shape(lhs, rhs)


class ElementwiseBinOpBase(IRDLOperation, ABC):
    """Base class for element-wise binary operations on tensors with Numpy-style broadcasting."""

    T = Annotated[AnyFloat | IntegerType, ConstraintVar("T")]
    lhs = operand_def(TensorType[T])
    rhs = operand_def(TensorType[T])
    res = result_def(TensorType[T])
    assembly_format = "`(` $lhs `,` $rhs `)` attr-dict `:` `(` type($lhs) `,` type($rhs) `)` `->` type($res)"

    def __init__(self, lhs: SSAValue, rhs: SSAValue, res_type: Attribute):
        super().__init__(
            operands=[lhs, rhs],
            result_types=[res_type],
        )

    def verify_(self) -> None:
        # Check that the arguments are broadcastable (using Numpy semantics) and that the result type is correct.
        if (
            not isinstance(lhs_type := self.lhs.type, TensorType)
            or not isinstance(rhs_type := self.rhs.type, TensorType)
            or not isinstance(res_type := self.res.type, TensorType)
        ):
            assert (
                False
            ), "onnx elementwise binary operation operands and result must be of type TensorType"
        lhs_type = cast(TensorType[Attribute], lhs_type)
        rhs_type = cast(TensorType[Attribute], rhs_type)
        res_type = cast(TensorType[Attribute], res_type)
        verify_multidirectional_broadcast_shape(lhs_type, rhs_type, res_type)


@irdl_op_definition
class Add(ElementwiseBinOpBase):
    name = "onnx.Add"


@irdl_op_definition
class Sub(ElementwiseBinOpBase):
    name = "onnx.Sub"


@irdl_op_definition
class Mul(ElementwiseBinOpBase):
    name = "onnx.Mul"


@irdl_op_definition
class Div(ElementwiseBinOpBase):
    name = "onnx.Div"


@irdl_op_definition
class Relu(IRDLOperation):
    """
    Relu takes one input data (Tensor) and produces one output data (Tensor) where the rectified linear function,
     y = max(0, x), is applied to the tensor elementwise.
    """

    name = "onnx.Relu"
    T = Annotated[AnyFloat | IntegerType, ConstraintVar("T")]
    operand = operand_def(TensorType[T])
    res = result_def(TensorType[T])
    assembly_format = (
        "`(` $operand`)` attr-dict `:` `(` type($operand) `)` `->` type($res)"
    )

    def __init__(self, operand: SSAValue):
        super().__init__(
            operands=[operand],
            result_types=[operand.type],
        )

    def verify_(self) -> None:
        assert isinstance(operand_type := self.operand.type, TensorType)
        assert isinstance(res_type := self.res.type, TensorType)
        operand_type = cast(TensorType[Attribute], operand_type)
        res_type = cast(TensorType[Attribute], res_type)

        if operand_type != res_type:
            raise VerifyException(
                "Mismatch between operand type and res type of onnx.Relu"
            )


@irdl_op_definition
class Gemm(IRDLOperation):
    """
    General Matrix multiplication: https://en.wikipedia.org/wiki/Basic_Linear_Algebra_Subprograms#Level_3
    A' = transpose(A) if transA else A
    B' = transpose(B) if transB else B
    Compute Y = alpha * A' * B' + beta * C,
    where input tensor A has shape (M, K) or (K, M), input tensor B has shape (K, N) or (N, K),
    input tensor C is broadcastable to shape (M, N), and output tensor Y has shape (M, N).
    """

    name = "onnx.Gemm"
    T = Annotated[AnyFloat | IntegerType, ConstraintVar("T")]
    tensor_a = operand_def(TensorType[T])
    tensor_b = operand_def(TensorType[T])
    tensor_c = operand_def(TensorType[T])

    alpha = opt_attr_def(FloatAttr[AnyFloat])
    beta = opt_attr_def(FloatAttr[AnyFloat])

    trans_a = opt_attr_def(AnyIntegerAttr, attr_name="transA")
    trans_b = opt_attr_def(AnyIntegerAttr, attr_name="transB")

    res_tensor = result_def(TensorType[T])
    assembly_format = (
        "`(` $tensor_a `,` $tensor_b `,`$tensor_c`)` attr-dict `:` `(` type($tensor_a) `,"
        "` type($tensor_b) `,`type($tensor_c)`)` `->` type($res_tensor) "
    )

    def __init__(
        self,
        tensor_a: SSAValue,
        tensor_b: SSAValue,
        tensor_c: SSAValue,
        alpha: Attribute,
        trans_a: Attribute,
        trans_b: Attribute,
        beta: Attribute,
    ):
        super().__init__(
            attributes={
                "transA": trans_a,
                "transB": trans_b,
                "alpha": alpha,
                "beta": beta,
            },
            operands=[tensor_a, tensor_b, tensor_c],
            result_types=[tensor_c.type],
        )

    def verify_(self) -> None:
        if (
            not isinstance(tensor_a_type := self.tensor_a.type, TensorType)
            or not isinstance(tensor_b_type := self.tensor_b.type, TensorType)
            or not isinstance(tensor_c_type := self.tensor_c.type, TensorType)
            or not isinstance(res_tensor_type := self.res_tensor.type, TensorType)
        ):
            assert (
                False
            ), "onnx elementwise operation operands and result must be of type TensorType"

        tensor_a_type = cast(TensorType[Attribute], tensor_a_type)
        tensor_b_type = cast(TensorType[Attribute], tensor_b_type)
        tensor_c_type = cast(TensorType[Attribute], tensor_c_type)
        res_tensor_type = cast(TensorType[Attribute], res_tensor_type)

        # store dimensions of tensor A and tensor B
        res_shape: list[int] = []
        tensor_a_shape = tensor_a_type.get_shape()
        tensor_b_shape = tensor_b_type.get_shape()

        if tensor_a_type.get_num_dims() != 2:
            raise VerifyException("tensor A should be a 2D tensor")

        if tensor_b_type.get_num_dims() != 2:
            raise VerifyException("tensor B should be a 2D tensor")

        if self.trans_a is not None and self.trans_a.value.data == 1:
            tensor_a_shape = tuple(reversed(tensor_a_shape))

        if self.trans_b is not None and self.trans_b.value.data == 1:
            tensor_b_shape = tuple(reversed(tensor_b_shape))

        if self.beta is not None:
            c_dims = tensor_c_type.get_num_dims()
            if c_dims > 2:
                raise VerifyException("tensor C should be a 1D tensor or 2D tensor")

        if tensor_a_shape[1] != tensor_b_shape[0]:
            raise VerifyException(
                f"operands have incompatible shapes: {tensor_a_shape} and {tensor_b_shape}"
            )
        else:
            res_shape.append(tensor_a_shape[0])
            res_shape.append(tensor_b_shape[1])

        # Build tensor of tensor (A * B) computation
        tensors_res = TensorType(tensor_a_type.element_type, res_shape)
        verify_unidirectional_broadcast_shape(
            tensors_res, tensor_c_type, res_tensor_type
        )


@irdl_op_definition
class Reshape(IRDLOperation):
    """
    Reshape the input tensor similar to numpy.reshape.
    First input is the data tensor, second input is a shape tensor which specifies the output shape. It outputs the reshaped tensor.
    At most one dimension of the new shape can be -1. In this case, the value is
    inferred from the size of the tensor and the remaining dimensions. A dimension
    could also be 0, in which case the actual dimension value is unchanged (i.e. taken
    from the input tensor). Shape (second input) could be an empty shape, which means converting to a scalar.
    The input tensor's shape and the output tensor's shape are required to have the same number of elements.

     Attributes:
        - allowzero  int (default is 0):  By default, when any value in the 'shape' input is equal to zero
     the corresponding dimension value is copied from the input tensor dynamically. allowzero=1 indicates that if any
     value in the 'shape' input is set to zero, the zero value is honoured, similar to NumPy.

    """

    name = "onnx.Reshape"
    T = Annotated[AnyFloat | IntegerType, ConstraintVar("T")]
    data = operand_def(TensorType[T])
    shape = operand_def(TensorType[IntegerType])
    reshaped = result_def(TensorType[T])

    allow_zero = opt_attr_def(AnyIntegerAttr, attr_name="allowzero")

    assembly_format = "`(` $data `,` $shape `)` attr-dict `:` `(` type($data) `,` type($shape) `)` `->` type($reshaped)"

    def __init__(self, data: SSAValue, shape: SSAValue, allow_zero: Attribute):
        super().__init__(
            attributes={"allowzero": allow_zero},
            operands=[data, shape],
            result_types=[data.type],
        )

    def verify_(self) -> None:
        if (
            not isinstance(data_type := self.data.type, TensorType)
            or not isinstance(shape_type := self.shape.type, TensorType)
            or not isinstance(reshaped_type := self.reshaped.type, TensorType)
        ):
            assert (
                False
            ), "onnx elementwise operation operands and result must be of type TensorType"

        data_type = cast(TensorType[Attribute], data_type)
        reshaped_type = cast(TensorType[Attribute], reshaped_type)
        shape_type = cast(TensorType[Attribute], shape_type)

        if shape_type.element_type != IntegerType(64):
            raise VerifyException(
                "shape element type has to be a 64-bit signless integer"
            )

        data_type = data_type.get_shape()
        shape_type = shape_type.get_shape()
        reshaped_type = reshaped_type.get_shape()

        # Shape tensor rank can't be -1.
        if shape_type[0] == -1:
            raise VerifyException("Shape tensor rank must not be equal to -1")

        # There is currently only support for rank one shape tensors in onnx-mlir
        # Shape tensor must have a constant shape.
        if len(shape_type) != 1:
            raise VerifyException("Shape tensor must have a rank one")

        # The input tensor's shape and the output tensor's shape are required to have the same number of elements.
        if math.prod(data_type) != math.prod(reshaped_type):
            raise VerifyException(
                "Input tensor's shape and output tensor's shape must have the same number of elements"
            )


@irdl_op_definition
class Abs(IRDLOperation):
    """
    Absolute takes one input data (Tensor) and produces one output data (Tensor) where absolute value,
    y = abs(x), is applied to the tensor elementwise.
    """

    name = "onnx.Abs"
    T = Annotated[AnyFloat | IntegerType, ConstraintVar("T")]
    operand = operand_def(TensorType[T])
    res = result_def(TensorType[T])
    assembly_format = (
        "`(` $operand`)` attr-dict `:` `(` type($operand) `)` `->` type($res)"
    )

    def __init__(self, operand: SSAValue):
        super().__init__(
            operands=[operand],
            result_types=[operand.type],
        )

    def verify_(self) -> None:
        assert isinstance(operand_type := self.operand.type, TensorType)
        assert isinstance(res_type := self.res.type, TensorType)
        operand_type = cast(TensorType[Attribute], operand_type)
        res_type = cast(TensorType[Attribute], res_type)
        if operand_type != res_type:
            raise VerifyException(
                "Mismatch between operand type and res type of onnx.Abs"
            )


@irdl_op_definition
class Conv(IRDLOperation):
    """
    The convolution operator consumes an input tensor and a filter, and computes the output.

    Attributes:

    - auto_pad  string (default is NOTSET): auto_pad must be either NOTSET, SAME_UPPER, SAME_LOWER or
    VALID. Where default value is NOTSET, which means explicit padding is used. SAME_UPPER or SAME_LOWER mean pad the
    input so that output_shape[i] = ceil(input_shape[i] / strides[i]) for each axis i. The padding is split between
    the two sides equally or almost equally (depending on whether it is even or odd). In case the padding is an odd
    number, the extra padding is added at the end for SAME_UPPER and at the beginning for SAME_LOWER.

    -dilations list of ints: dilation value along each spatial axis of the filter. If not present, the dilation
    defaults is 1 along each spatial axis.

    -group int (default is '1'): number of groups input channels and output channels are divided into.

    -kernel_shape list of ints: The shape of the convolution kernel. If not present, should be inferred from input W.

    -pads list of ints: Padding for the beginning and ending along each spatial axis, it can take any value greater
    than or equal to 0. The value represent the number of pixels added to the beginning and end part of the
    corresponding axis. `pads` format should be as follow [x1_begin, x2_begin...x1_end, x2_end,...], where xi_begin
    the number of pixels added at the beginning of axis `i` and xi_end, the number of pixels added at the end of axis
    `i`. This attribute cannot be used simultaneously with auto_pad attribute. If not present, the padding defaults
    to 0 along start and end of each spatial axis.

    -strides list of ints: Stride along each spatial axis. If not present, the stride defaults is 1 along each spatial axis.

    """

    name = "onnx.Conv"
    T = Annotated[AnyFloat | IntegerType, ConstraintVar("T")]
    data = operand_def(TensorType[T])
    weight = operand_def(TensorType[T])
    bias = operand_def(TensorType[T] | NoneType)
    res = result_def(TensorType[T])

    auto_pad = attr_def(StringAttr)
    dilations = attr_def(ArrayAttr[AnyIntegerAttr])
    group = attr_def(AnyIntegerAttr)
    kernel_shape = attr_def(ArrayAttr[AnyIntegerAttr])
    pads = attr_def(ArrayAttr[AnyIntegerAttr])
    strides = attr_def(ArrayAttr[AnyIntegerAttr])

    assembly_format = (
        "`(` $data `,` $weight `,`$bias`)` attr-dict `:` `(` type($data) `,"
        "` type($weight) `,`type($bias)`)` `->` type($res) "
    )

    def __init__(
        self,
        data: SSAValue,
        weight: SSAValue,
        bias: SSAValue,
        auto_pad: Attribute,
        dilations: Attribute,
        group: Attribute,
        kernel_shape: Attribute,
        pads: Attribute,
        strides: Attribute,
    ):
        super().__init__(
            attributes={
                "auto_pad": auto_pad,
                "dilations": dilations,
                "group": group,
                "kernel_shape": kernel_shape,
                "pads": pads,
                "strides": strides,
            },
            operands=[data, weight, bias],
            result_types=[data.type],
        )

    def verify_(self) -> None:
        if (
            not isinstance(data_type := self.data.type, TensorType)
            or not isinstance(weight_type := self.weight.type, TensorType)
            or not isinstance(bias_type := self.bias.type, TensorType | NoneType)
            or not isinstance(res_type := self.res.type, TensorType)
        ):
            assert False, (
                "onnx elementwise operation operands (data, weight) and result (res) must be of type TensorType,"
                "operand (bias) must be of type TensorType or NoneType"
            )

        weight_type = cast(TensorType[Attribute], weight_type)
        data_type = cast(TensorType[Attribute], data_type)
        res_type = cast(TensorType[Attribute], res_type)

        # case that bias is a tensor type
        if isinstance(bias_type, TensorType):
            bias_type = bias_type.get_shape()
            if len(bias_type) != 1:
                raise VerifyException("bias must be 1D")

        weight_type = weight_type.get_shape()
        # kernel_shape
        kernel_shape_data: list[int] = []
        for value in self.kernel_shape:
            val = value.value.data
            kernel_shape_data.append(val)
        if list(weight_type[-2:]) != kernel_shape_data:
            raise VerifyException(
                "kernel shape rank and weight tensor rank are not the same"
            )

        # dilations
        for value in self.dilations:
            val = value.value.data
            if val <= 0:
                raise VerifyException("dilation value must be non zero positive")
        if len(self.dilations) != len(self.kernel_shape):
            raise VerifyException(
                "dilations rank and kernel shape rank are not the same"
            )

        # group
        if self.group.value.data < 1:
            raise VerifyException("group value must be nonnegative")

        # strides
        for value in self.strides:
            val = value.value.data
            if val <= 0:
                raise VerifyException("stride value must be non zero positive")
        if len(self.strides) != len(self.kernel_shape):
            raise VerifyException(
                "strides rank and kernel shape rank are not the same "
            )

        # pads
        for value in self.pads:
            val = value.value.data
            if val < 0:
                raise VerifyException("pads value must be nonnegative")
        if len(self.pads) != 2 * len(self.kernel_shape):
            raise VerifyException("pads rank is not twice the kernel shape rank")

        auto_pad_strings = ["NOTSET", "SAME_UPPER", "SAME_LOWER", "VALID"]
        if self.auto_pad.data not in auto_pad_strings:
            raise VerifyException(
                f"Invalid auto_pad string. Must be one of {auto_pad_strings}"
            )


@irdl_op_definition
class Constant(IRDLOperation):
    """
    Produce a constant tensor.

    Exactly one of the provided attributes, either value, sparse_value, or value_* must be specified.

    Attributes:
    - sparse_value: sparse_tensor
    The value for the elements of the output tensor in sparse format. (currently unsupported)
    - value : tensor
     The value for the elements of the output tensor.
    - value_float: float
     The value for the sole element for the scalar, float32, output tensor.
    - value_floats: list of floats
     The values for the elements for the 1D, float32, output tensor.
    - value_int : int
     The value for the sole element for the scalar, int64, output tensor.
    - value_ints : list of ints
     The values for the elements for the 1D, int64, output tensor.
    - value_string : string
     The value for the sole element for the scalar, UTF-8 string, output tensor.
    - value_strings: list of strings
     The values for the elements for the 1D, UTF-8 string, output tensor.
    """

    name = "onnx.Constant"
    output = result_def(AnyTensorType)

    value = opt_attr_def(DenseIntOrFPElementsAttr)
    value_float = opt_attr_def(FloatAttr[Float32Type])
    value_floats = opt_attr_def(ArrayAttr[FloatAttr[Float32Type]])
    value_int = opt_attr_def(IntegerAttr[IntegerType])
    value_ints = opt_attr_def(ArrayAttr[IntegerAttr[IntegerType]])
    value_string = opt_attr_def(StringAttr)
    value_strings = opt_attr_def(ArrayAttr[StringAttr])

    def __init__(
        self,
        value: Attribute | None,
        value_float: Attribute | None,
        value_floats: Attribute | None,
        value_int: Attribute | None,
        value_ints: Attribute | None,
        value_string: Attribute | None,
        value_strings: Attribute | None,
        output_type: Attribute | None,
    ):
        super().__init__(
            attributes={
                "value": value,
                "value_float": value_float,
                "value_floats": value_floats,
                "value_int": value_int,
                "value_ints": value_ints,
                "value_string": value_string,
                "value_strings": value_strings,
            },
            operands=[],
            result_types=[output_type],
        )

    def verify_(self) -> None:
        if self.value is not None and not isinstance(self.value.type, TensorType):
            raise VerifyException("value attribute type must be of type TensorType")

        if self.value_int is not None and self.value_int.type.width.data != 64:
            raise VerifyException(
                "value_int element type has to be a 64-bit signless integer"
            )

        if self.value_ints is not None:
            for value in self.value_ints:
                width = value.type.width.data
                if width != 64:
                    raise VerifyException(
                        "value_ints elements type has to be a 64-bit signless integer"
                    )

        attrs = [
            self.value,
            self.value_float,
            self.value_floats,
            self.value_int,
            self.value_ints,
            self.value_string,
            self.value_strings,
        ]
        used_attrs = sum(1 for attr in attrs if attr is not None)
        if used_attrs != 1:
            raise VerifyException(
                f"Only one value attribute must be provided, but {used_attrs} were specified"
            )

    def print(self, printer: Printer):
        if self.value is not None:
            printer.print(" ")
            printer.print(self.value)

    @classmethod
    def parse(cls, parser: Parser) -> Self:
        v = parser.parse_attribute()
        if not isinstance(v, DenseIntOrFPElementsAttr):
            raise NotImplementedError()
        constant = cls(v, None, None, None, None, None, None, v.type)
        return constant


@irdl_op_definition
class MaxPoolSingleOut(IRDLOperation):
    """
    ONNX MaxPool operation with a single output.

     Attributes:

    - auto_pad string (default is 'NOTSET'):  auto_pad must be either NOTSET, SAME_UPPER, SAME_LOWER or
    VALID. Where default value is NOTSET, which means explicit padding is used. SAME_UPPER or SAME_LOWER mean pad the
    input so that output_shape[i] = ceil(input_shape[i] / strides[i]) for each axis i. The padding is split between
    the two sides equally or almost equally (depending on whether it is even or odd). In case the padding is an odd
    number, the extra padding is added at the end for SAME_UPPER and at the beginning for SAME_LOWER.

    - ceil_mode  int (default is '1'): Whether to use ceil or floor (default) to compute the output shape.

    - dilations list of ints: Dilation value along each spatial axis of filter.

    - kernel_shape list of ints: The size of the kernel along each axis.

    - pads list of ints: Padding for the beginning and ending along each spatial axis, it can take any value greater
    than or equal to 0. The value represent the number of pixels added to the beginning and end part of the
    corresponding axis. `pads` format should be as follow [x1_begin, x2_begin...x1_end, x2_end,...], where xi_begin
    the number of pixels added at the beginning of axis `i` and xi_end, the number of pixels added at the end of axis
    `i`. This attribute cannot be used simultaneously with auto_pad attribute. If not present, the padding defaults
    to 0 along start and end of each spatial axis.

    - storage_order int (default is '0') : The storage order of the tensor. 0 is row major, and 1 is column major.
    This attribute is used only to convert an n-tuple index value into a single integer value for producing the
    second output.

    - strides list of ints: Stride along each spatial axis. If not present, the stride defaults to 1 along each spatial axis

    """

    name = "onnx.MaxPoolSingleOut"

    T = Annotated[AnyFloat | IntegerType, ConstraintVar("T")]
    data = operand_def(TensorType[T] | MemRefType[T])
    output = result_def(TensorType[T] | MemRefType[T])

    auto_pad = attr_def(StringAttr)
    ceil_mode = attr_def(AnyIntegerAttr)
    dilations = attr_def(ArrayAttr[AnyIntegerAttr])
    kernel_shape = attr_def(ArrayAttr[AnyIntegerAttr])
    pads = attr_def(ArrayAttr[AnyIntegerAttr])
    storage_order = attr_def(AnyIntegerAttr)
    strides = attr_def(ArrayAttr[AnyIntegerAttr])

    assembly_format = (
        "`(` $data`)` attr-dict `:` `(` type($data) `)` `->` type($output)"
    )

    def __init__(
        self,
        data: SSAValue,
        auto_pad: Attribute,
        ceil_mode: Attribute,
        dilations: Attribute,
        kernel_shape: Attribute,
        pads: Attribute,
        storage_order: Attribute,
        strides: Attribute,
    ):
        super().__init__(
            attributes={
                "auto_pad": auto_pad,
                "ceil_mode": ceil_mode,
                "dilations": dilations,
                "kernel_shape": kernel_shape,
                "pads": pads,
                "storage_order": storage_order,
                "strides": strides,
            },
            operands=[data],
            result_types=[data.type],
        )

    def verify_(self) -> None:
        if not isinstance(
            data_type := self.data.type, TensorType | MemRefType
        ) or not isinstance(output_type := self.output.type, TensorType | MemRefType):
            assert False, (
                "onnx elementwise operation operands (data) and result (output) must be of type TensorType or "
                "MemRefTyoe "
            )

        data_type = cast(TensorType[Attribute], data_type)
        output_type = cast(TensorType[Attribute], output_type)

        # auto pad
        auto_pad_strings = ["NOTSET", "SAME_UPPER", "SAME_LOWER", "VALID"]
        if self.auto_pad.data not in auto_pad_strings:
            raise VerifyException(
                f"Invalid auto_pad string. Must be one of {auto_pad_strings}"
            )

        # ceil mode
        if self.ceil_mode.value.data < 0 or self.ceil_mode.value.data > 1:
            raise VerifyException("ceil value must be either zero or one")

        # kernel shape
        if (input_dims := len(data_type.get_shape()) - 2) != (
            kernel_dims := len(self.kernel_shape)
        ):
            raise VerifyException(
                f"input data and kernel shape rank mismatch: ({input_dims}) vs ({kernel_dims})"
            )

        # dilations
        for value in self.dilations:
            val = value.value.data
            if val <= 0:
                raise VerifyException("dilation value must be non zero positive")

        if (dilations_dims := len(self.dilations)) != (
            kernel_dims := len(self.kernel_shape)
        ):
            raise VerifyException(
                f"dilations rank ({dilations_dims}) and kernel shape rank ({kernel_dims}) are not the "
                f"same "
            )

        # storage order
        # Not supported for storage order in column major mode in onnx-mlir (therefore row major mode only considered)
        if self.storage_order.value.data != 0:
            raise VerifyException("column major storage order not implemented yet")

        # strides
        for value in self.strides:
            val = value.value.data
            if val <= 0:
                raise VerifyException("stride value must be non zero positive")

        if (strides_dims := len(self.strides)) != (
            kernel_dims := len(self.kernel_shape)
        ):
            raise VerifyException(
                f"strides rank ({strides_dims}) and kernel shape rank ({kernel_dims}) are not the "
                f"same "
            )

        # pads
        for value in self.pads:
            val = value.value.data
            if val < 0:
                raise VerifyException("pads value must be nonnegative")

        if (pads_dims := len(self.pads)) != 2 * len(self.kernel_shape):
            raise VerifyException(
                f"pads rank ({pads_dims}) is not twice the kernel shape rank ({len(self.kernel_shape)})"
            )


@irdl_op_definition
class EntryPoint(IRDLOperation):
    """
    Indicate ONNX entry point
    The "onnx.EntryPoint" function indicates the main entry point of ONNX model.
    """

    name = "onnx.EntryPoint"
    func = attr_def(SymbolRefAttr)

    def __init__(self, func: Attribute):
        super().__init__(
            attributes={
                "func": func,
            },
        )


@irdl_op_definition
class MatMul(IRDLOperation):
    """
    The operation MatMul performs matrix multiplication between two input matrices, A and B, and returns the result as matrix Y.
    Matrix multiplication is a fundamental operation in linear algebra, where each element of the resulting matrix Y is computed by taking the
    dot product of the corresponding row of matrix A and column of matrix B.
    """

    name = "onnx.MatMul"

    # describe annotated type
    T = Annotated[AnyFloat | IntegerType, ConstraintVar("T")]

    # input matrices
    matrix_A = operand_def(TensorType[T])
    matrix_B = operand_def(TensorType[T])

    # output matrices
    matrix_Y = result_def(TensorType[T])

    assembly_format = (
        "`(` $matrix_A `,` $matrix_B `)` attr-dict `:` `(` type($matrix_A) `,"
        "` type($matrix_B) `)` `->` type($matrix_Y) "
    )

    def __init__(
        self,
        matrix_A: SSAValue,
        matrix_B: SSAValue,
        matrix_Y_type: Attribute,
    ):
        super().__init__(
            operands=[matrix_A, matrix_B],
            result_types=[matrix_Y_type],
        )

    def verify_(self) -> None:
        # store dimensions of tensor A and tensor B
        res_shape: list[int] = []
        matrix_A_type = cast(TensorType[Any], self.matrix_A.type)
        matrix_B_type = cast(TensorType[Any], self.matrix_B.type)
        matrix_Y_type = cast(TensorType[Any], self.matrix_Y.type)

        # check shape compatibility
        matrix_A_shape = matrix_A_type.get_shape()
        matrix_B_shape = matrix_B_type.get_shape()

        if matrix_A_type.get_num_dims() != 2:
            raise VerifyException("input matrix A should be a 2D tensor")

        if matrix_B_type.get_num_dims() != 2:
            raise VerifyException("input matrix B should be a 2D tensor")

        if matrix_A_shape[1] != matrix_B_shape[0]:
            raise VerifyException(
                f"operands have incompatible shapes: {matrix_A_shape} and {matrix_B_shape}"
            )
        else:
            res_shape.append(matrix_A_shape[0])
            res_shape.append(matrix_B_shape[1])

        matrix_Y_type_shape = list(matrix_Y_type.get_shape())
        if (
            len(res_shape) != len(matrix_Y_type_shape)
            or res_shape != matrix_Y_type_shape
        ):
            raise VerifyException(
                f"result shape {res_shape} does not match result type {matrix_Y_type_shape}"
            )


@irdl_op_definition
class Transpose(IRDLOperation):
    """
    The transpose_tensor function takes a tensor as input and returns its transpose.
    Transposing a tensor means flipping its dimensions, so that rows become columns and vice versa.
    """

    name = "onnx.Transpose"

    T = Annotated[AnyFloat | IntegerType, ConstraintVar("T")]
    tensor_input = operand_def(TensorType[T])

    perm = opt_attr_def(ArrayAttr[AnyIntegerAttr], attr_name="perm")

    tensor_output = result_def(TensorType[T])

    assembly_format = (
        "`(` $tensor_input `)` attr-dict `:` `(` type($tensor_input) "
        "`)` `->` type($tensor_output) "
    )

    def __init__(self, tensor_input: SSAValue, perm: Attribute):
        super().__init__(
            attributes={"perm": perm},
            operands=[tensor_input],
            result_types=[tensor_input.type],
        )

    def verify_(self) -> None:
        if not isinstance(
            tensor_input_type := self.tensor_input.type, TensorType
        ) or not isinstance(tensor_output_type := self.tensor_output.type, TensorType):
            assert (
                False
            ), "onnx elementwise operation operands and result must be of type TensorType"

        tensor_input_shape = tensor_input_type.get_shape()
        tensor_output_shape = tensor_output_type.get_shape()

        # numbers in perm cannot be repeated
        if self.perm is not None:

            for _, int_attr in enumerate(self.perm.data):
                attr_value = int_attr.value.data
                count = self.perm.data.count(int_attr)
                if count != 1:
                    raise VerifyException(
                        f"permutation can not contain more than one occurrence of the same dimension: dimension #{attr_value} appears {count} times."
                    )

            # numbers in perm must be between 0 and len(tensor_input_shape)-1
            perm_size = len(self.perm.data)
            for int_index, int_attr in enumerate(self.perm.data):
                int_index = int_index + 0
                int_attr_val = int_attr.value.data
                if int_attr_val < 0 or int_attr_val >= perm_size:
                    raise VerifyException(
                        f"permutation can only contain values between 0 and {perm_size}-1: dimension #{int_index} value is {int_attr_val}"
                    )

            # len(tensor_input_shape) must be equal to len(perm)
            perm_size = len(self.perm.data)
            input_size = len(tensor_input_shape)
            if perm_size != input_size:
                raise VerifyException(
                    f"permutation and inputs dimensions must have the same size: #dimensions input is {input_size}, #dimension perimutation is {perm_size}"
                )

            # check output shape
            for index_attr, int_attr in enumerate(self.perm.data):
                int_attr_val = int_attr.value.data
                if tensor_output_shape[index_attr] != tensor_input_shape[int_attr_val]:
                    raise VerifyException(
                        f"incorrect output shape: output dimension #{index_attr} should be equal to {tensor_input_shape[int_attr_val]}"
                    )


@irdl_op_definition
class Squeeze(IRDLOperation):
    """
    Squeeze the input tensor along the specified axes.

    Squeezing a tensor removes dimensions of size 1, effectively reducing the rank of the tensor and collapsing those dimensions.
    This operation is particularly useful for removing unnecessary singleton dimensions, which may arise from broadcasting or previous operations.

    Args:
        input_tensor: The input tensor to be squeezed. This tensor should be a multi-dimensional array-like object.
        axes: A list of axes along which to squeeze the tensor. If provided, only the specified axes will be squeezed. If not provided, all dimensions of size 1 will be squeezed.

    Returns:
        output_tensor: The squeezed tensor.
    """

    name = "onnx.Squeeze"

    T = Annotated[AnyFloat | IntegerType, ConstraintVar("T")]
    input_tensor = operand_def(TensorType[T])
    axes = opt_attr_def(IntegerAttr, attr_name="axes")

    output_tensor = result_def(TensorType[T])

    assembly_format = "`(` $input_tensor `)` attr-dict `:` `(` type($input_tensor) `)` `->` type($output_tensor) "

    def __init__(
        self,
        input_tensor: SSAValue,
        axes: Attribute,
    ):
        super().__init__(
            attributes={
                "axes": axes,
            },
            operands=[input_tensor],
            result_types=[input_tensor.type],
        )

    def verify_(self) -> None:
        if not isinstance(input_tensor_type := self.input_tensor.type, TensorType):
            assert (
                False
            ), "onnx elementwise operation operands and result must be of type TensorType"

        input_tensor_shape = input_tensor_type.get_shape()

        if self.axes is not None:
            axes_value = self.axes.value.data

            # axes out of bounds: the axes value must between 0 and len(input_tensor.shape)-1
            if axes_value < 0 or axes_value >= len(input_tensor_shape):
                max_axes_value = len(input_tensor_shape) - 1
                raise VerifyException(
                    f"axes to squeeze must be between 0 and {max_axes_value}, axes: {axes_value}"
                )


@irdl_op_definition
class Sigmoid(IRDLOperation):
    """
    Applies the sigmoid function element-wise to all elements of the input tensor.
    The sigmoid function, denoted by sigma(x), is a common mathematical function used in machine learning and neural networks. It is defined as:
    sigma(x) = 1 / (1 + e^-x)
    where e is the base of the natural logarithm. The sigmoid function maps any real-valued number to the range of [0, 1].
    The sigmoid function is used as an activation function.

    Args:
    - input_tensor (TensorType): The input tensor to which the sigmoid function will be applied.

    Returns:
    - output_tensor (TensorType): The output tensor after applying the sigmoid function element-wise to the input tensor.
    """

    name = "onnx.Sigmoid"

    T = Annotated[AnyFloat, ConstraintVar("T")]
    input_tensor = operand_def(TensorType[T])
    output_tensor = result_def(TensorType[T])

    assembly_format = "`(` $input_tensor`)` attr-dict `:` `(` type($input_tensor) `)` `->` type($output_tensor) "

    def __init__(
        self,
        input_tensor: SSAValue,
    ):
        super().__init__(
            operands=[input_tensor],
            result_types=[input_tensor.type],
        )

    def verify_(self) -> None:
        if not isinstance(
            input_tensor_type := self.input_tensor.type, TensorType
        ) or not isinstance(output_tensor_type := self.output_tensor.type, TensorType):
            assert (
                False
            ), "onnx elementwise operation operands and result must be of type TensorType"

        input_tensor_shape = input_tensor_type.get_shape()
        output_tensor_shape = output_tensor_type.get_shape()

        # check if input tensor and output tensor have the same shape
        if input_tensor_shape != output_tensor_shape:
            raise VerifyException(
                f"tensor input shape {input_tensor_shape} is not equal to tensor output shape {output_tensor_shape}"
            )


ONNX = Dialect(
    "onnx",
    [
        Abs,
        Add,
        Constant,
        Conv,
        Div,
        EntryPoint,
        Gemm,
        MatMul,
        MaxPoolSingleOut,
        Mul,
        Relu,
        Reshape,
        Sub,
        Transpose,
        Squeeze,
        Sigmoid,
    ],
)
